from __future__ import annotations

from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper_common.TypeDesc import TypeDesc

test_data_ids_to_load_ = "test_data_ids_to_load"


class ServiceLoaderConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    # List of `test_data` ids which are allowed to load.
    # Those `data_envelopes` which do not have `test_data` id assigned are loaded regardless of any filter.
    test_data_ids_to_load = fields.List(
        fields.String(),
        required = True,
    )


service_loader_config_desc = TypeDesc(
    dict_schema = ServiceLoaderConfigSchema(),
    ref_name = ServiceLoaderConfigSchema.__name__,
    dict_example = {
        test_data_ids_to_load_: [
            "TD_70_69_38_46",  # no data
        ],
    },
    default_file_path = "",
)



from __future__ import annotations

from typing import Union

from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay.enum_desc.ResultCategory import ResultCategory
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.schema_config_core_client.ClientConfigSchema import client_config_desc


class PluginCheckEnvServerUrl(PluginCheckEnvAbstract):
    """
    Simple plugin which lists URL-s for configured servers.
    """

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        check_env_results: list[CheckEnvResult] = []
        client_config: ClientConfig = client_config_desc.obj_from_default_file()
        for server_index, redundant_server in enumerate(client_config.redundant_servers):
            check_env_results.append(CheckEnvResult(
                result_category = ResultCategory.VerificationSuccess,
                result_key = f"server_url[{server_index}]",
                result_value = f"http://{redundant_server.server_host_name}:{redundant_server.server_port_number}",
                result_message = None,
            ))
        return check_env_results



"""
Functions to work with FS_97_64_39_94 `arg_bucket`-related data.
"""

from __future__ import annotations


def arg_buckets_to_token_ipos_list(
    arg_buckets: list[list[int]],
) -> list[int]:
    """
    Converts `arg_bucket`-s to a flat token ipos list.
    """
    token_ipos_list = []
    for arg_bucket in arg_buckets:
        for token_ipos in arg_bucket:
            token_ipos_list.append(token_ipos)
    return token_ipos_list



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_context.arg_buckets_utils import arg_buckets_to_token_ipos_list
from argrelay.schema_response.ArgValues import ArgValues


@dataclass
class InterpResult(ArgValues):
    """
    See also :class:`InterpResultSchema`.

    Unlike `InvocationInput` this class contains at most one `data_envelope` per `envelope_container`
    because it executes in the mode which queries values only for latency-sensitive Tab-completion.
    """

    all_tokens: list[str] = field()
    """
    Copy from `ParsedContext.all_tokens` - command line args.
    """

    excluded_tokens: list[int] = field()
    """
    Copy from `InterpContext.excluded_tokens` -
    indexes into `all_tokens` pointing to tokens with special meaning (e.g. tangent token, arg bucket separator, ...).
    """

    consumed_arg_buckets: list[list[int]] = field()
    """
    Copy from `InterpContext.consumed_arg_buckets` -
    indexes into `all_tokens` pointing to tokens consumed during interpretation.
    """

    envelope_containers: list[EnvelopeContainer] = field()

    tan_token_ipos: int = field()
    """
    Value from `ParsedContext.tan_token_ipos`.
    """

    tan_token_l_part: str = field()
    """
    Value from `ParsedContext.tan_token_l_part`.
    """

    @staticmethod
    def from_interp_context(
        interp_ctx: InterpContext,
    ) -> InterpResult:
        return InterpResult(
            arg_values = interp_ctx.comp_suggestions,
            all_tokens = interp_ctx.parsed_ctx.all_tokens,
            excluded_tokens = interp_ctx.excluded_tokens,
            consumed_arg_buckets = interp_ctx.consumed_arg_buckets,
            envelope_containers = interp_ctx.envelope_containers,
            tan_token_ipos = interp_ctx.parsed_ctx.tan_token_ipos,
            tan_token_l_part = interp_ctx.parsed_ctx.tan_token_l_part,
        )

    def get_data_envelopes(self) -> list[dict]:
        data_envelopes = []
        for envelope_container in self.envelope_containers:
            data_envelopes.extend(envelope_container.data_envelopes)
        return data_envelopes

    def consumed_token_ipos_list(
        self,
    ) -> list[int]:
        return arg_buckets_to_token_ipos_list(self.consumed_arg_buckets)



from __future__ import annotations

from typing import Union

from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.handler_response.ClientResponseHandlerAbstract import ClientResponseHandlerAbstract
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_client.ClientCommandAbstract import ClientCommandAbstract
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.server_spec.CallContext import CallContext


class ClientCommandLocal(ClientCommandAbstract):

    def __init__(
        self,
        call_ctx: CallContext,
        client_response_handler: ClientResponseHandlerAbstract,
        local_server: LocalServer,
        server_request_handler: AbstractServerRequestHandler,
    ):
        super().__init__(
            call_ctx,
        )
        self.client_response_handler: ClientResponseHandlerAbstract = client_response_handler
        self.local_server: LocalServer = local_server
        self.server_request_handler: AbstractServerRequestHandler = server_request_handler
        self.response_dict: Union[dict, None] = None
        self.interp_ctx: Union[InterpContext, None] = None

    def execute_command(
        self,
    ):
        self.response_dict = self.server_request_handler.handle_request(self.call_ctx)
        ElapsedTime.measure("before_sending_response")
        self.client_response_handler.handle_response(self.response_dict)
        ElapsedTime.measure("after_handle_response")

        # `ClientCommandLocal`'s feature: expose `InterpContext` used by the server:
        self.interp_ctx = self.server_request_handler.interp_ctx



class ServerResponseError(Exception):
    """
    Exception for failed server responses.

    It is related to `ClientExitCode.ServerError`.

    Specifically, it is NOT connection error.
    """

    pass


def print_full_stack_trace(given_exception):
    """
    Print full stack trace equivalent to not catching exception.

    Somehow Python does not have standard function
    (only stack trace to current line and stack trace for exception from current caller)
    which requires implementations like these:
    https://stackoverflow.com/a/16589622/441652
    """
    import traceback, sys
    exception_obj = sys.exc_info()[0]
    # Remove curr func frame:
    caller_stack = traceback.extract_stack()[:-1]
    if exception_obj is not None:
        # Remove curr func frame:
        del caller_stack[-1]
    head_message = "Traceback (most recent call last):\n"
    caller_stack_str = head_message + "".join(traceback.format_list(caller_stack))
    if exception_obj is not None:
        exception_stack_str = traceback.format_exc().lstrip(head_message)
        full_stack_str = caller_stack_str + exception_stack_str
    else:
        exc_line = f"{type(given_exception).__name__}: {str(given_exception)}"
        full_stack_str = caller_stack_str + exc_line
    print(full_stack_str, file = sys.stderr)



from __future__ import annotations

from argrelay.composite_forest.CompositeNodeSchema import func_id_
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
)
from argrelay.schema_config_interp.SearchControlSchema import populate_search_control
from argrelay.schema_response.InvocationInput import InvocationInput


class NoopDelegator(AbstractDelegator):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        class_to_collection_map: dict = self.server_config.class_to_collection_map

        no_data_search_control = populate_search_control(
            class_to_collection_map,
            "no_data_class",
            [
                {"no_data_key_one": "no_data_prop_one"},
                {"no_data_key_two": "no_data_prop_two"},
            ],
        )

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: SpecialFunc.func_id_unplugged.name,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Function which does nothing and is not supposed to be plugged in",
                # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
                # Tagged as `FuncState.fs_ignorable` but, if unplugged, it becomes `FuncState.func_id_unplugged`:
                ReservedPropName.func_state.name: FuncState.fs_ignorable.name,
                ReservedPropName.func_id.name: SpecialFunc.func_id_unplugged.name,
            },
            {
                instance_data_: {
                    func_id_: SpecialFunc.func_id_no_data.name,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        no_data_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Function which has no data it searches for and does nothing",
                ReservedPropName.func_state.name: FuncState.fs_ignorable.name,
                ReservedPropName.func_id.name: SpecialFunc.func_id_no_data.name,
            },
        ]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.plugin_instance_entries[
                self.plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        # Do nothing:
        pass



from __future__ import annotations

from typing import Union

import argrelay
from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay.enum_desc.ResultCategory import ResultCategory


class PluginCheckEnvModuleVersion(PluginCheckEnvAbstract):
    """
    Trivial plugin to report `argrelay.__version__`.
    """

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        return [CheckEnvResult(
            result_category = ResultCategory.VerificationSuccess,
            result_key = "argrelay_module_version",
            result_value = argrelay.__version__,
            result_message = None,
        )]



from __future__ import annotations

from enum import Enum


class TopDir(Enum):
    """
    Top-level directories for `argrelay` - see also `argrelay_dir`.

    See also FS_29_54_67_86 project `dir_structure`.
    """

    bin_dir = "bin"

    conf_dir = "conf"

    data_dir = "data"

    docs_dir = "docs"

    dst_dir = "dst"

    exe_dir = "exe"

    src_dir = "src"

    test_dir = "test"

    tmp_dir = "tmp"

    var_dir = "var"

    def __str__(self):
        return self.name



from enum import Enum


class ProcRole(Enum):
    """
    Defines role of the (operating system) process.

    Part of FS_14_59_14_06 pending requests implementation.
    """

    SoleProcWorker = False, True
    """
    Combination of `ParentProcSpinner` and `ChildProcWorker` within single process.
    """

    ParentProcSpinner = True, False
    """
    A parent (foreground) process which spins waiting for data from `ChildProcWorker`.
    """

    ChildProcWorker = True, True
    """
    A child (background) process which makes request to the server and sends data to `ParentProcSpinner`.
    """

    CheckEnvWorker = False, True
    """
    Similar to `SoleProcWorker` but used exclusively for FS_36_17_84_44 `check_env` implementation.
    """

    def __new__(
        cls,
        is_split_mode: bool,
        is_worker_proc: bool,
    ):
        enum_value = len(cls.__members__) + 1
        enum_obj = object.__new__(cls)
        enum_obj._value_ = enum_value
        return enum_obj

    def __init__(
        self,
        is_split_mode: bool,
        is_worker_proc: bool,
    ):
        self.is_split_mode = is_split_mode
        self.is_worker_proc = is_worker_proc



from __future__ import annotations

from typing import Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.PluginType import PluginType
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.runtime_context.AbstractPluginServer import AbstractPluginServer
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.InterpContext import function_container_ipos_
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc
from argrelay.schema_response.InvocationInput import InvocationInput


def get_func_id_from_interp_ctx(
    interp_ctx: "InterpContext",
):
    """
    Used on the server-side of the plugin.
    """

    func_data_envelope = interp_ctx.envelope_containers[(
        interp_ctx.curr_interp.base_container_ipos + function_container_ipos_
    )].data_envelopes[0]
    return func_data_envelope[instance_data_][func_id_]


def get_func_id_from_invocation_input(
    invocation_input: InvocationInput,
):
    """
    Used on the client-side of the plugin.
    """

    func_data_envelope = invocation_input.envelope_containers[
        function_container_ipos_
    ].data_envelopes[0]
    return func_data_envelope[instance_data_][func_id_]


class AbstractDelegator(AbstractPluginServer):
    """
    `DelegatorPlugin` implements two sides:
    *   server-side `invoke_control` prepares data in :class:`InvocationInput` (whatever is necessary)
    *   client-side `invoke_action` uses data in :class:`InvocationInput` to execute the action anyway it can

    To simplify reasoning, ensure that both server-side and client-side:
    *   have access to the same code (non-breaking differences are possible, but same code version ensures that)
    *   share data only via :class:`InvocationInput`
    """

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.DelegatorPlugin

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        """
        Part of FS_26_43_73_72 func tree implementation.
        """
        return []

    def run_search_control(
        self,
        function_data_envelope: dict,
    ) -> list[SearchControl]:
        """
        Implements FS_31_70_49_15 `search_control`.
        """
        return self.extract_search_control_from_function_data_envelope(function_data_envelope)

    def run_init_control(
        self,
        envelope_containers: list[EnvelopeContainer],
        curr_container_ipos: int,
    ) -> None:
        """
        Implements FS_46_96_59_05 `init_control`.
        """
        self.init_envelope_class(
            envelope_containers,
            curr_container_ipos,
        )

    def has_fill_control(
        self,
    ) -> bool:
        """
        FS_72_53_55_13: Optimization for showing values hidden by defaults.

        *   If returns True, `run_fill_control` will be called subsequently
            (and one more query will be run with populated values).
        *   If returns False, `run_fill_control` will not be called
            (and extra query is not necessary since there were no change of values by `run_fill_control`).
        """
        return False

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ) -> bool:
        """
        Implements FS_72_40_53_00 `fill_control`.

        FS_72_53_55_13: Optimization for showing values hidden by defaults:
        *   If returns True, it means at least one value was populated = one more query is necessary.
        *   If returns False, no values were changed = extra query with populated values is not necessary.
        """
        return False

    def run_interp_control(
        self,
        curr_interp: "AbstractInterp",
    ) -> Union[None, str]:
        """
        Implements FS_78_91_27_22 `interp_control`.

        Selects `plugin_id` for the next interp.
        """
        pass

    def run_invoke_control(
        self,
        interp_ctx: "InterpContext",
        local_server: "LocalServer",
    ) -> InvocationInput:
        """
        Implements FS_98_55_40_77 `invoke_control` on server side.
        The plugin instance is used by server on `ServerAction.RelayLineArgs`.
        """
        pass

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        """
        Implements FS_98_55_40_77 `invoke_control` on client side.
        There is no plugin instance on client side -
        instead, the class is used directly (statically) on `ServerAction.RelayLineArgs`:
        *   without instantiating a plugin (it is a `@staticmethod`)
        *   without calling `AbstractPlugin.activate_plugin`
        """
        pass

    @staticmethod
    def extract_search_control_from_function_data_envelope(
        function_data_envelope: dict,
    ) -> list[SearchControl]:
        return [
            search_control_desc.dict_schema.load(search_control_dict)
            for search_control_dict in function_data_envelope[instance_data_][search_control_list_]
        ]

    @staticmethod
    def init_envelope_class(
        envelope_containers: list[EnvelopeContainer],
        curr_container_ipos: int,
    ) -> None:
        """
        Sets `ReservedPropName.envelope_class` according to FS_31_70_49_15 search control.
        """
        curr_container = envelope_containers[curr_container_ipos]
        curr_container.assigned_types_to_values[
            ReservedPropName.envelope_class.name
        ] = AssignedValue(
            curr_container.search_control.envelope_class,
            ArgSource.InitValue,
        )



from typing import Type

from argrelay.handler_response.ClientResponseHandlerAbstract import ClientResponseHandlerAbstract
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.runtime_context.AbstractPlugin import import_plugin_class
from argrelay.schema_response.InvocationInput import InvocationInput
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc


class ClientResponseHandlerRelayLineArgs(ClientResponseHandlerAbstract):

    def __init__(
        self,
    ):
        super().__init__(
        )

    def handle_response(
        self,
        response_dict: dict,
    ):
        invocation_input: InvocationInput = invocation_input_desc.dict_schema.load(response_dict)
        ElapsedTime.measure("after_object_creation")
        plugin_class: Type[AbstractDelegator] = import_plugin_class(invocation_input.delegator_plugin_entry)
        plugin_class.invoke_action(invocation_input)



from enum import Enum, auto


class ServiceEnvelopeClass(Enum):
    ClassCluster = auto()
    ClassHost = auto()
    ClassService = auto()
    ClassAccessType = auto()

    def __str__(self):
        return self.name



from __future__ import annotations

from copy import deepcopy
from typing import Union

from argrelay.composite_forest.CompositeForestExtractor import extract_jump_tree, extract_func_tree
from argrelay.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay.composite_forest.DictTreeWalker import DictTreeWalker, normalize_tree, sequence_starts_with
from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.misc_helper_server import insert_unique_to_sorted_list
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory, AbstractInterp
from argrelay.plugin_interp.FuncTreeInterpFactoryConfigSchema import (
    func_tree_interp_config_desc,
)
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.InitControl import InitControl
from argrelay.runtime_context.InterpContext import InterpContext, function_container_ipos_
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import delegator_plugin_instance_id_
from argrelay.schema_config_interp.InitControlSchema import init_types_to_values_, init_control_desc
from argrelay.schema_config_interp.SearchControlSchema import (
    keys_to_types_list_,
    populate_search_control,
    search_control_desc,
)
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_enabled_, plugin_dependencies_

func_search_control_ = "func_search_control"
"""
This field is automatically populated by `FuncTreeInterpFactory` inside `interp_tree_node_config_dict`.
"""

func_init_control_ = "func_init_control"
"""
This field is automatically populated by `FuncTreeInterpFactory` inside `interp_tree_node_config_dict`.
"""

tree_path_selector_prefix_ = "tree_path_selector_"
tree_path_selector_key_prefix_ = "s"


class FuncTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_26_43_73_72 func tree.
    Implements FS_91_88_07_23 jump tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        # FS_26_43_73_72 func tree: func id to list of its absolute paths populated by `load_func_envelopes`.
        # These func tree paths are absolute within FS_33_76_82_84 composite forest.
        # Each func id can be attached to more than one leaf (hence, there is a list of paths to that func id).
        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = {}

        # FS_91_88_07_23 jump tree
        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.jump_tree,
            extract_jump_tree(
                self.server_config.server_plugin_control.composite_forest,
            ),
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = dict_tree_walker.build_paths_to_paths()
        """
        Implements FS_91_88_07_23 jump tree:
        for given `interp_tree_abs_path` (FS_01_89_09_24) selects next `interp_tree_abs_path`.
        """

        self.plugin_search_control: Union[SearchControl, None] = None

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return func_tree_interp_config_desc.dict_from_input_dict(plugin_config_dict)

    def provide_plugin_search_control(
        self,
    ) -> list[SearchControl]:
        if self.plugin_search_control:
            return [self.plugin_search_control]
        else:
            return []

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        self.interp_tree_abs_paths.extend(
            this_plugin_instance_interp_tree_abs_paths,
        )

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        """
        To implement FS_26_43_73_72 func tree, this plugin loads func `data_envelope`-s automatically.

        It loops through func `data_envelope`-s and populates their tree path search props
        according to extracted `func_tree` where each function is plugged into.
        """

        mapped_func_ids: list[str] = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        interp_tree_node_config_dict = self.interp_tree_abs_paths_to_node_configs[interp_tree_abs_path]

        func_selector_tree = normalize_tree(extract_func_tree(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_instance_id,
        ))

        dict_tree_walker = DictTreeWalker(
            CompositeInfoType.func_tree,
            func_selector_tree,
        )
        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()

        # `func_envelope` (2-level map) per `func_id` per `func_abs_path`
        func_id_to_func_abs_path_to_func_envelope: dict[str, dict[tuple[str, ...], dict]] = {}
        # Loop through func `data_envelope`-s from all delegators:
        for func_id, func_envelope in func_ids_to_func_envelopes.items():
            if func_id not in self.func_ids_to_func_abs_paths:
                # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
                # This `func_id` is not plugged into given tree.
                # It will be loaded as `FuncState.fs_unplugged`.
                continue
            func_abs_paths = self.func_ids_to_func_abs_paths[func_id]
            for func_abs_path in func_abs_paths:
                assert (
                    func_id not in func_id_to_func_abs_path_to_func_envelope
                    or
                    tuple(func_abs_path) not in func_id_to_func_abs_path_to_func_envelope[func_id]
                )
                if self._is_best_matching_abs_tree_path(func_abs_path, interp_tree_abs_path):
                    # This `func_id` is mapped into the `func_tree` under this plugin:
                    func_id_to_func_abs_path_to_func_envelope.setdefault(
                        func_id,
                        {},
                    )[tuple(func_abs_path)] = deepcopy(func_envelope)
                    mapped_func_ids.append(func_id)

        # Ensure every `func_id` mapped into the tree is part of `func_envelope`-s loaded from delegators:
        for func_id in self.func_ids_to_func_abs_paths.keys():
            if func_id not in mapped_func_ids:
                raise RuntimeError(
                    f"plugin_instance_id=`{self.plugin_instance_id}`: func_id=`{func_id}` is not published by any enabled delegator activated so far - please check `{plugin_enabled_}` and `{plugin_dependencies_}`"
                )

        self.populate_func_envelope_collection(
            func_id_to_func_abs_path_to_func_envelope,
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )

        return mapped_func_ids

    def populate_func_envelope_collection(
        self,
        func_id_to_func_abs_path_to_func_envelope: dict[str, dict[tuple[str, ...], dict]],
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
    ):
        self.populate_init_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )
        self.populate_search_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )
        (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        ) = self.populate_func_tree_props(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
            func_id_to_func_abs_path_to_func_envelope,
        )
        class_to_collection_map: dict = self.server_config.class_to_collection_map
        class_names = [
            ReservedEnvelopeClass.ClassFunction.name,
        ]
        init_envelop_collections(
            self.server_config,
            class_names,
            lambda collection_name, class_name: (
                [
                    ReservedPropName.envelope_class.name,
                ] + prop_names
            )
        )
        envelope_collection = self.server_config.static_data.envelope_collections[
            class_to_collection_map[ReservedEnvelopeClass.ClassFunction.name]
        ]
        # Write func envelopes into `StaticData` (as if it is a loader plugin):
        envelope_collection.data_envelopes.extend(interp_tree_abs_path_func_envelopes)

    def _is_best_matching_abs_tree_path(
        self,
        func_abs_path,
        given_interp_tree_abs_path,
    ) -> bool:
        if sequence_starts_with(func_abs_path, given_interp_tree_abs_path):
            best_matching_abs_tree_path = given_interp_tree_abs_path
            max_matched_len = len(best_matching_abs_tree_path)
            for interp_tree_abs_path in self.interp_tree_abs_paths:
                if given_interp_tree_abs_path == interp_tree_abs_path:
                    continue
                if sequence_starts_with(func_abs_path, interp_tree_abs_path):
                    if max_matched_len < len(interp_tree_abs_path):
                        best_matching_abs_tree_path = interp_tree_abs_path
                        max_matched_len = len(best_matching_abs_tree_path)
                        break
                    elif max_matched_len == len(interp_tree_abs_path):
                        # There should not be two plugin paths leading to the same func:
                        raise RuntimeError()
                    else:
                        continue
            return given_interp_tree_abs_path == best_matching_abs_tree_path
        else:
            return False

    # noinspection PyMethodMayBeStatic
    def populate_init_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict: dict,
    ):
        interp_tree_node_config_dict[func_init_control_] = {
            init_types_to_values_: {},
        }

        init_types_to_values = interp_tree_node_config_dict[func_init_control_][init_types_to_values_]
        for sel_ipos, path_step in enumerate(interp_tree_abs_path):
            init_types_to_values[f"{func_envelope_path_step_prop_name(sel_ipos)}"] = path_step

    def populate_search_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
    ):
        """
        Provide `func_search_control` based on `func_ids_to_func_abs_paths`.
        """

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        plugin_search_control_dict: dict = populate_search_control(
            class_to_collection_map,
            ReservedEnvelopeClass.ClassFunction.name,
            [],
        )

        # `func_search_control` should include keys from the interp tree abs path:
        keys_to_types_list = plugin_search_control_dict[keys_to_types_list_]

        # Include func tree path:
        max_len = max_path_len(self.func_ids_to_func_abs_paths)
        for sel_ipos in range(max_len):
            keys_to_types_list.append({
                f"{path_step_key_arg_name(sel_ipos)}": f"{func_envelope_path_step_prop_name(sel_ipos)}"
            })

        # Include other fields:
        keys_to_types_list.append({
            "state": ReservedPropName.func_state.name
        })
        keys_to_types_list.append({
            "id": ReservedPropName.func_id.name
        })

        interp_tree_node_config_dict[func_search_control_] = plugin_search_control_dict
        self.plugin_search_control = search_control_desc.obj_from_input_dict(plugin_search_control_dict)

    def populate_func_tree_props(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
        func_envelopes_index: dict[str, dict[tuple[str, ...], dict]],
    ) -> tuple[list[dict], list[str]]:
        """
        Populates func tree properties for each func `data_envelope` based on its path in func tree.

        For each func envelope:
        *   Select its place on the func tree (provided by plugin config).
        *   Populate the path as envelope props with `tree_path_selector_prefix_`.
        """
        interp_tree_abs_path_func_envelopes: list[dict] = []
        prop_names: list[str] = []
        for func_id in self.func_ids_to_func_abs_paths:
            func_abs_paths: list[list[str]] = self.func_ids_to_func_abs_paths[func_id]

            for func_abs_path in func_abs_paths:

                if (
                    func_id not in func_envelopes_index
                    or
                    tuple(func_abs_path) not in func_envelopes_index[func_id]
                ):
                    continue

                func_envelope = func_envelopes_index[func_id][tuple(func_abs_path)]

                # Include func tree path:
                for sel_ipos, path_step_id in enumerate(func_abs_path):
                    prop_name = func_envelope_path_step_prop_name(sel_ipos)
                    func_envelope[prop_name] = path_step_id
                    prop_names.append(prop_name)

                interp_tree_abs_path_func_envelopes.append(func_envelope)

        return (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        return FuncTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[interp_ctx.interp_tree_abs_path],
            interp_ctx,
            self.func_ids_to_func_abs_paths,
            self.paths_to_jump,
        )


def func_envelope_path_step_prop_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_path_selector_prefix_}{path_step_ipos}"


def path_step_key_arg_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_path_selector_key_prefix_}{path_step_ipos}"


def max_path_len(
    func_ids_to_func_abs_paths: dict[str, list[list[str]]],
):
    """
    Find maximum path length.
    """
    max_len: int = 0
    for func_rel_paths in func_ids_to_func_abs_paths.values():
        for func_rel_path in func_rel_paths:
            max_len = max(max_len, len(func_rel_path))
    return max_len


class FuncTreeInterp(AbstractInterp):
    """
    Implements FS_26_43_73_72 func tree.

    Finds function `data_envelope` within func tree first,
    then uses its delegator (see `AbstractDelegator`) to find all args-related `data_envelope`-s.

    See FS_55_57_45_04 enum selector.
    """

    def __init__(
        self,
        interp_factory_id,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        func_ids_to_func_abs_paths: dict[str, list[list[str]]],
        paths_to_jump: dict[tuple[str, ...], tuple[str, ...]],
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = paths_to_jump

        # Allocate first container for function `data_envelope`:
        self.base_container_ipos += 1
        self.interp_ctx.envelope_containers.append(EnvelopeContainer(SearchControl()))

        self.select_next_container()
        self._apply_func_init_control()
        self._apply_func_search_control()

        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = func_ids_to_func_abs_paths

    def _apply_func_init_control(self):
        self.interp_ctx.curr_container.assigned_types_to_values[
            ReservedPropName.envelope_class.name
        ] = AssignedValue(
            ReservedEnvelopeClass.ClassFunction.name,
            ArgSource.InitValue,
        )
        func_init_control: InitControl = init_control_desc.dict_schema.load(
            self.interp_tree_node_config_dict[func_init_control_],
        )
        for prop_type, prop_value in func_init_control.init_types_to_values.items():
            self.interp_ctx.curr_container.assigned_types_to_values[prop_type] = AssignedValue(
                prop_value,
                ArgSource.InitValue,
            )
            if prop_type in self.interp_ctx.curr_container.remaining_types_to_values:
                del self.interp_ctx.curr_container.remaining_types_to_values[prop_type]

    def _apply_func_search_control(self):
        # Function `search_control` is populated based on
        # tree path (FS_01_89_09_24 interp tree + FS_26_43_73_72 func tree)
        # and plugin config (rather than data found in `data_envelope`):
        self.interp_ctx.curr_container.search_control = search_control_desc.dict_schema.load(
            self.interp_tree_node_config_dict[func_search_control_]
        )

    def consume_pos_args(self) -> bool:
        """
        Scans through `remaining_arg_buckets` and tries to match its value against values of each type.

        Implements:
        *   FS_76_29_13_28 arg consumption priorities.
        *   FS_44_36_84_88 consume args one by one:
            This func consumes all until the first remaining non-singled out arg.
        *   FS_97_64_39_94 `arg_bucket`-s: consumption is limited to single bucket per `envelope_container`.
        """

        consumed_token_ipos_list = []
        any_consumed = False
        # Related to FS_13_51_07_97 singled out implicit values:
        # We can keep consuming args (without creating FS_51_67_38_37 impossible arg combinations)
        # as long as they are singled out - we cannot consume two ambiguous args at once, but
        # we can consume as many singled out as possible (plus one ambiguous).
        # If arg is singled out but still matches remaining arg, it must be assigned as `ArgSource.ExplicitPosArg`
        # rather than be left remaining and (later) be assigned as `ArgSource.ImplicitValue`.
        consumed_ambiguous_value = False
        if self.interp_ctx.curr_container.used_arg_bucket is not None:
            # If `envelope_container` has one `used_arg_bucket`, loop through it only:
            any_consumed = self.consume_pos_args_from_arg_bucket(
                bucket_index = self.interp_ctx.curr_container.used_arg_bucket,
                bucket_list = self.interp_ctx.remaining_arg_buckets[self.interp_ctx.curr_container.used_arg_bucket],
                consumed_ambiguous_value = consumed_ambiguous_value,
                consumed_token_ipos_list = consumed_token_ipos_list,
            )
        else:
            # Otherwise, loop through all buckets until the single `used_arg_bucket` is chosen:
            for bucket_index, bucket_list in enumerate(self.interp_ctx.remaining_arg_buckets):
                any_consumed = self.consume_pos_args_from_arg_bucket(
                    bucket_index = bucket_index,
                    bucket_list = bucket_list,
                    consumed_ambiguous_value = consumed_ambiguous_value,
                    consumed_token_ipos_list = consumed_token_ipos_list,
                )
                if any_consumed:
                    # Consume from single `arg_bucket` only:
                    break

        # perform list modifications out of the prev loop:
        for consumed_token_ipos in consumed_token_ipos_list:
            bucket_index = self.interp_ctx.token_ipos_to_arg_bucket_map[consumed_token_ipos]
            self.interp_ctx.remaining_arg_buckets[bucket_index].remove(consumed_token_ipos)

        return any_consumed

    def consume_pos_args_from_arg_bucket(
        self,
        bucket_index,
        bucket_list,
        consumed_ambiguous_value,
        consumed_token_ipos_list,
    ) -> bool:
        any_consumed = False
        for remaining_token_ipos in bucket_list:

            remaining_token = self.interp_ctx.parsed_ctx.all_tokens[remaining_token_ipos]

            # TODO: FS_76_29_13_28 Why not define the order based on FS_31_70_49_15 `search_control`
            #       (instead of whatever internal order `remaining_types_to_values` has)?
            #       It could already be the case that `remaining_types_to_values` are ordered as `search_control`.
            #       Why not make it explicit?

            # See if token matches any type by value:
            for arg_type, arg_values in self.interp_ctx.curr_container.remaining_types_to_values.items():
                if remaining_token in arg_values:
                    if (
                        len(arg_values) == 1
                        or
                        not consumed_ambiguous_value
                    ):
                        self.interp_ctx.curr_container.assigned_types_to_values[arg_type] = AssignedValue(
                            remaining_token,
                            ArgSource.ExplicitPosArg,
                        )
                        self.interp_ctx.curr_container.used_arg_bucket = bucket_index
                        if len(arg_values) > 1:
                            # This was not singled out arg:
                            # allow only one ambiguous consumption to avoid FS_51_67_38_37 impossible arg combinations.
                            consumed_ambiguous_value = True
                        any_consumed = True
                        consumed_token_ipos_list.append(remaining_token_ipos)

                        self.interp_ctx.consumed_arg_buckets[bucket_index].append(remaining_token_ipos)

                        # TD_76_09_29_31: overlapped
                        # Assign matching remaining arg value to the first type it matches (only once):
                        del self.interp_ctx.curr_container.remaining_types_to_values[arg_type]
                        break
        return any_consumed

    def try_iterate(self) -> InterpStep:
        """
        Try to consume more args if possible.

        *   If function was found, start with search for its first envelope class.
        *   If curr envelope class is found, move to the next (until all are found).

        :returns:
        *   `InterpStep.NextInterp`: move to next interpreter: curr interpreter is fully satisfied from the args
        *   `InterpStep.NextEnvelope`: call again curr interpreter: still more things to find in the args
        *   `InterpStep.StopAll`: interpreter sees no point to continue the loop (`InterpContext.interpret_command`)
        """

        # We want single `data_envelope` to be found, not zero, not more than one:
        if self.interp_ctx.curr_container.found_count > 1:
            # Too many `data_envelope`-s - stop:
            return InterpStep.StopAll
        elif self.interp_ctx.curr_container.found_count == 1:

            if self.interp_ctx.curr_container_ipos == self.base_container_ipos:
                # This is a function envelope:
                search_control_list: list[SearchControl] = self.get_search_control_list()
                # Create `EnvelopeContainer`-s for every `data_envelope` to find:
                self.interp_ctx.alloc_searchable_containers(search_control_list)

            if self.interp_ctx.is_last_container():
                # Function does not need any envelopes:
                return InterpStep.NextInterp
            else:
                self.select_next_container()
                self.delegate_init_control()
                # Need more args to consume for the next envelope to find:
                return InterpStep.NextEnvelope

        else:
            # No `data_envelope` = nothing to do:
            return InterpStep.StopAll

    def get_search_control_list(self) -> list[SearchControl]:
        delegator_plugin = self.get_func_delegator()

        search_control_list: list[SearchControl] = delegator_plugin.run_search_control(
            self.get_found_func_data_envelope()
        )
        return search_control_list

    def delegate_init_control(self):
        delegator_plugin = self.get_func_delegator()
        delegator_plugin.run_init_control(
            self.interp_ctx.envelope_containers,
            self.interp_ctx.curr_container_ipos,
        )

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def delegate_fill_control(
        self,
    ) -> bool:
        delegator_plugin = self.get_func_delegator()
        if delegator_plugin:
            return delegator_plugin.run_fill_control(
                self.interp_ctx,
            )
        else:
            return False

    def get_found_func_data_envelope(self) -> Union[dict, None]:
        func_envelope = self.interp_ctx.envelope_containers[(
            self.base_container_ipos + function_container_ipos_
        )]
        if func_envelope.found_count == 1:
            return func_envelope.data_envelopes[0]
        return None

    def get_func_delegator(self):
        func_data_envelope = self.get_found_func_data_envelope()
        if func_data_envelope:
            delegator_plugin_instance_id = func_data_envelope[instance_data_][delegator_plugin_instance_id_]
            delegator_plugin: AbstractDelegator = self.interp_ctx.action_delegators[delegator_plugin_instance_id]
            return delegator_plugin
        else:
            # func envelope hasn't been found yet:
            return None

    def select_next_container(self):
        self.interp_ctx.curr_container_ipos += 1
        self.interp_ctx.curr_container = self.interp_ctx.envelope_containers[
            self.interp_ctx.curr_container_ipos
        ]

    def next_interp(self) -> "AbstractInterp":
        delegator_plugin = self.get_func_delegator()
        interp_factory_id = delegator_plugin.run_interp_control(self)
        if interp_factory_id:
            self.select_next_interp_tree_abs_path()
            return self.interp_ctx.create_next_interp(interp_factory_id)
        else:
            return None

    def select_next_interp_tree_abs_path(self):
        if self.interp_ctx.interp_tree_abs_path in self.paths_to_jump:
            # FS_91_88_07_23 jump tree: replace current `interp_tree_abs_path` with another one based on config:
            self.interp_ctx.interp_tree_abs_path = self.paths_to_jump[self.interp_ctx.interp_tree_abs_path]

    def propose_arg_completion(self) -> None:
        for comp_value in self.propose_auto_comp_list():
            insert_unique_to_sorted_list(self.interp_ctx.comp_suggestions, comp_value)

    def propose_auto_comp_list(self) -> list[str]:

        # TODO: FS_20_88_05_60: POC: Either remove it or implement properly: just testing named args:
        if (
            self.interp_ctx.parsed_ctx.tan_token_l_part.endswith(":")
            or
            self.interp_ctx.parsed_ctx.tan_token_r_part.startswith(":")
        ):
            return [
                type_name + SpecialChar.KeyValueDelimiter.value
                for type_name in self.interp_ctx.curr_container.search_control.types_to_keys_dict
                if not type_name.startswith("_")
            ]

        # TODO: FS_23_62_89_43: the logic for both if-s (`if-A` and `if-B`) is identical at the moment - what do we want to improve?

        # TODO: FS_23_62_89_43: if-A:
        if self.interp_ctx.parsed_ctx.comp_scope is CompScope.ScopeInitial:
            if self.interp_ctx.parsed_ctx.tan_token_l_part == "":
                return self.remaining_from_next_missing_type()
            else:
                return self.remaining_from_next_missing_type()

        # TODO: FS_23_62_89_43: if-B:
        if self.interp_ctx.parsed_ctx.comp_scope is CompScope.ScopeSubsequent:
            if self.interp_ctx.parsed_ctx.tan_token_l_part == "":
                return self.remaining_from_next_missing_type()
            else:
                # TODO: FS_20_88_05_60: Suggest keys (:) of missing types instead - it is `ScopeSubsequent`, user insist and wants something else:
                return self.remaining_from_next_missing_type()

        return []

    def remaining_from_next_missing_type(self) -> list[str]:
        """
        Clarifications:
        *   remaining = because values for the given type are reduced based on narrowed down `data_envelope` set
        *   missing = because this arg type is not specified yet
        *   next = because arg types are tired in specific order
        """
        proposed_values: list[str] = []

        # Return filtered value set from the next missing arg:
        for arg_type in self.interp_ctx.curr_container.search_control.types_to_keys_dict:
            if (
                # TODO: only one condition should be enough: arg_type is either in one or in another, not in both:
                arg_type not in self.interp_ctx.curr_container.assigned_types_to_values
                and
                arg_type in self.interp_ctx.curr_container.remaining_types_to_values
            ):
                proposed_values = [
                    # FS_71_87_33_52: `help_hint`:
                    self.interp_ctx.help_hint_cache.get_value_with_help_hint(arg_type, x)
                    for x in self.interp_ctx.curr_container.remaining_types_to_values[arg_type]
                    if (
                        # NOTE: This checks for `str` meaning primitive (scalar, not `list` or `dict`),
                        #       but this also filters out `int` or others.
                        #       Primitive types should be converted to `str` when loaded.
                        #       See `ConfigOnlyLoader.convert_envelope_fields_to_string` for example.
                        isinstance(x, str)
                        and
                        # FS_32_05_46_00: using `startswith`:
                        # FS_23_62_89_43: filter using L part of tangent token:
                        x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                        # TODO: FS_06_99_43_60: Support list[str] - what if one type can have list of values (and we need to match any as in OR)?
                    )
                ]
                if proposed_values:
                    # Collect only until the first proposed value set from missing args:
                    break

        return proposed_values



from argrelay.client_pipeline.BytesHandlerAbstract import BytesHandlerAbstract


class BytesHandlerTextProposeArgValuesOptimized(BytesHandlerAbstract):

    def __init__(
        self,
    ):
        super().__init__(
        )

    def handle_bytes(
        self,
        given_bytes: bytes,
    ):
        # For (default) "text/plain" response, proposed (new-line-separated) values are directly in the body:
        print(given_bytes.decode())



from marshmallow import Schema, RAISE

from argrelay.misc_helper_common.TypeDesc import TypeDesc


# TODO_40_10_18_32: add custom base to all schemas:
class FuncTreeInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True


func_tree_interp_config_example = {
}

func_tree_interp_config_desc = TypeDesc(
    dict_schema = FuncTreeInterpFactoryConfigSchema(),
    ref_name = FuncTreeInterpFactoryConfigSchema.__name__,
    dict_example = func_tree_interp_config_example,
    default_file_path = "",
)



from marshmallow import fields, RAISE

from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.relay_server.UsageStatsEntry import UsageStatsEntry
from argrelay.schema_request.CallContextSchema import (
    server_action_,
    command_line_,
    call_context_desc,
    comp_scope_,
    cursor_cpos_,
)

server_ts_ns_ = "server_ts_ns"
client_version_ = "client_version"
client_conf_target_ = "client_conf_target"
client_user_id_ = "client_user_id"


class UsageStatsEntrySchema(ObjectSchema):
    """
    Schema for FS_87_02_77_34: usage stats entry - see `UsageStatsEntry`.
    """

    class Meta:
        unknown = RAISE
        ordered = True

    model_class = UsageStatsEntry

    server_action = fields.Enum(
        ServerAction,
        required = True,
    )

    comp_scope = fields.Enum(
        CompScope,
        required = True,
    )

    server_ts_ns = fields.Integer(
        required = False,
        load_default = 0,
    )

    client_version = fields.String(
        required = False,
        load_default = "",
    )

    client_conf_target = fields.String(
        required = False,
        load_default = "",
    )

    client_user_id = fields.String(
        required = False,
        load_default = "",
    )

    command_line = fields.String(
        required = True,
    )

    cursor_cpos = fields.Integer(
        required = True,
    )


usage_stats_entry_desc = TypeDesc(
    dict_schema = UsageStatsEntrySchema(),
    ref_name = UsageStatsEntrySchema.__name__,
    dict_example = {
        server_action_: call_context_desc.dict_example[server_action_],
        comp_scope_: call_context_desc.dict_example[comp_scope_],
        server_ts_ns_: 0,
        client_version_: "",
        client_conf_target_: "",
        client_user_id_: "",
        command_line_: call_context_desc.dict_example[command_line_],
        cursor_cpos_: call_context_desc.dict_example[cursor_cpos_],
    },
    default_file_path = "",
)



from __future__ import annotations

from marshmallow import RAISE, fields

from argrelay.composite_forest.CompositeForestSchema import composite_forest_desc
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.ServerPluginControl import ServerPluginControl

first_interp_factory_id_ = "first_interp_factory_id"
composite_forest_ = "composite_forest"


class ServerPluginControlSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = ServerPluginControl

    first_interp_factory_id = fields.String(
        required = True,
    )

    composite_forest = fields.Nested(
        composite_forest_desc.dict_schema,
        required = True,
    )


server_plugin_control_desc = TypeDesc(
    dict_schema = ServerPluginControlSchema(),
    ref_name = ServerPluginControlSchema.__name__,
    dict_example = {
        first_interp_factory_id_: "SomeInterp",
        composite_forest_: composite_forest_desc.dict_example,
    },
    default_file_path = "",
)



from __future__ import annotations

from json import dumps

from flasgger import swag_from
from flask import request, Blueprint, Response, abort

from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.handler_request.DescribeLineArgsServerRequestHandler import DescribeLineArgsServerRequestHandler
from argrelay.handler_request.ProposeArgValuesServerRequestHandler import ProposeArgValuesServerRequestHandler
from argrelay.handler_request.RelayLineArgsServerRequestHandler import RelayLineArgsServerRequestHandler
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.schema_request.CallContextSchema import call_context_desc
from argrelay.schema_response.ArgValuesSchema import arg_values_
from argrelay.server_spec import DescribeLineArgsSpec, ProposeArgValuesSpec, RelayLineArgsSpec
from argrelay.server_spec.CallContext import CallContext


def create_blueprint_api(local_server: LocalServer):
    blueprint_api = Blueprint("blueprint_api", __name__)

    describe_line_args_handler = DescribeLineArgsServerRequestHandler(local_server)
    propose_arg_values_handler = ProposeArgValuesServerRequestHandler(local_server)
    relay_line_args_handler = RelayLineArgsServerRequestHandler(local_server)

    def _create_call_ctx() -> CallContext:
        ElapsedTime.clear_measurements()
        ElapsedTime.measure("before_request_payload_load")

        call_ctx: CallContext
        if request.content_type is None:
            data_json = request.data.decode("utf-8")
            call_ctx = call_context_desc.dict_schema.loads(data_json)
        elif request.content_type == "application/json":
            call_ctx = call_context_desc.dict_schema.load(request.json)
        else:
            abort(415)

        ElapsedTime.measure("after_call_context_creation")
        ElapsedTime.is_debug_enabled = call_ctx.is_debug_enabled
        return call_ctx

    @blueprint_api.route(
        ServerAction.ProposeArgValues.value,
        methods = ["post"],
    )
    @swag_from(ProposeArgValuesSpec.spec_data)
    def propose_arg_values():
        try:
            call_ctx = _create_call_ctx()
            response_dict = propose_arg_values_handler.handle_request(call_ctx)

            if request.accept_mimetypes["text/plain"] or len(request.accept_mimetypes) == 0:
                # Required for `ClientCommandRemoteWorkerTextProposeArgValuesOptimized`:
                # Sending "text/plain" (default) for stripped down clients (who may not even specify headers)
                # also serving perf reasons on client side (trivial parsing, no lib required, minimal imports):
                return Response(
                    "\n".join(response_dict[arg_values_]),
                    mimetype = "text/plain",
                )
            elif request.accept_mimetypes["application/json"]:
                # JSON - parsing lib is required on the client side:
                response_json = dumps(response_dict)
                return Response(
                    response_json,
                    mimetype = "application/json",
                )
            else:
                # Not acceptable:
                abort(406)
        finally:
            ElapsedTime.measure("before_sending_response")

    @blueprint_api.route(
        ServerAction.DescribeLineArgs.value,
        methods = ["post"],
    )
    @swag_from(DescribeLineArgsSpec.spec_data)
    def describe_line_args():
        try:
            call_ctx = _create_call_ctx()
            response_dict = describe_line_args_handler.handle_request(call_ctx)

            if request.accept_mimetypes["application/json"] or len(request.accept_mimetypes) == 0:
                response_json = dumps(response_dict)
                return Response(
                    response_json,
                    mimetype = "application/json",
                )
            else:
                # Not acceptable:
                abort(406)
        finally:
            ElapsedTime.measure("before_sending_response")

    @blueprint_api.route(
        ServerAction.RelayLineArgs.value,
        methods = ["post"],
    )
    @swag_from(RelayLineArgsSpec.spec_data)
    def relay_line_args():
        try:
            call_ctx = _create_call_ctx()
            response_dict = relay_line_args_handler.handle_request(call_ctx)

            if request.accept_mimetypes["application/json"] or len(request.accept_mimetypes) == 0:
                response_json = dumps(response_dict)
                return Response(
                    response_json,
                    mimetype = "application/json",
                )
            else:
                # Not acceptable:
                abort(406)
        finally:
            ElapsedTime.measure("before_sending_response")

    @blueprint_api.teardown_request
    def show_teardown(exception):
        ElapsedTime.print_all_if_debug()

    return blueprint_api



from enum import IntEnum


class ClientExitCode(IntEnum):
    """
    Lists argrelay client exit codes.
    """

    ClientSuccess = 0

    GeneralError = 1

    ConnectionError = 2
    """
    Used to communicate specifically connection issues between client `ProcRole`-s.

    Provides a mechanism to implement FS_93_18_57_91 client fail over.
    """

    ServerError = 3
    """
    Server responded with an error (e.g. HTTP status code not 200),
    but it is not a `ConnectionError`.

    See also `ServerResponseError`.
    """

    # maximum exit code:
    # https://stackoverflow.com/a/67219932/441652
    UnknownError = 255



from __future__ import annotations

import dataclasses
import re
from dataclasses import dataclass, field

from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper_common import eprint
from argrelay.server_spec.CallContext import CallContext


@dataclass(frozen = True)
class ParsedContext(CallContext):
    """
    Internal immutable parsed view of :class:`InputContext`

    Implements `FS_27_16_67_19` line syntax.

    `tan_token_*` = `tangent_token`, see FS_23_62_89_43.
    """

    all_tokens: list[str] = field(init = False)
    tan_token_ipos: int = field(init = False)
    tan_token_l_cpos: int = field(init = False)
    tan_token_r_cpos: int = field(init = False)
    tan_token: str = field(init = False)
    tan_token_l_part: str = field(init = False)
    tan_token_r_part: str = field(init = False)

    def __post_init__(self):
        (
            all_tokens,
            tan_token_ipos,
            tan_token_l_cpos,
            tan_token_r_cpos,
            tan_token,
            tan_token_l_part,
            tan_token_r_part,
        ) = self.parse_input(self)
        object.__setattr__(self, 'all_tokens', all_tokens)
        object.__setattr__(self, 'tan_token_ipos', tan_token_ipos)
        object.__setattr__(self, 'tan_token_l_cpos', tan_token_l_cpos)
        object.__setattr__(self, 'tan_token_r_cpos', tan_token_r_cpos)
        object.__setattr__(self, 'tan_token', tan_token)
        object.__setattr__(self, 'tan_token_l_part', tan_token_l_part)
        object.__setattr__(self, 'tan_token_r_part', tan_token_r_part)

    @classmethod
    def from_instance(
        cls,
        call_ctx: CallContext,
    ):
        return cls(**dataclasses.asdict(call_ctx))

    @staticmethod
    def parse_input(
        call_ctx: CallContext,
    ) -> tuple[
        list[str],
        int,
        int,
        int,
        str,
        str,
        str,
    ]:
        """
        Given `|` is the cursor in this command line:
        ```
        some_command some_su|b_command some_arg % next_arg   last_arg
        ```
        ```
                            |                                         # non-existing char placed to indicate cursor cpos
                                                %                     # arg bucket separator
        0            1                 2        3 4          5        # token ipos
        01234567890123456789 0123456789012345678901234567890123456789 # char cpos (the least significant digit)
        ```
        Then function returns:
        (
            [                               # list of all tokens
                "some_command",             # 0
                "some_sub_command",         # 1 (tangent token = token "touched" by the cursor)
                "some_arg",                 # 2
                "%",                        # 3 arg bucket separator
                "next_arg",                 # 4
                "last_arg"                  # 5
            ],
            1                               # tangent token item position (ipos)
            13                              # left-most char position (cpos) of tangent token (including it)
            29                              # right-most char position (cpos) of tangent token (excluding it)
            "some_sub_command",             # tangent token string
            "some_su",                      # tangent token left part substring (preceding cursor position)
            "b_command",                    # tangent token right part substring (succeeding cursor position)
        )
        """
        # Wrap orig command line into delimiter for simplification:
        command_line = f" {call_ctx.command_line} "
        cursor_cpos = call_ctx.cursor_cpos + 1
        line_len = len(command_line)

        # Init with defaults:
        # TODO: FS_23_62_89_43: Tangent token ipos should always be above zero and point to
        #                       (possibly surrogate missing empty) token ipos within command line.
        all_tokens = []
        tan_token_ipos = -1
        tan_token_l_cpos = -1
        tan_token_r_cpos = -1

        # Iterate through all delimiter spans to cut out tokens:
        prev_span_l = -1
        prev_span_r = -1
        token_ipos = 0
        for m in re.finditer(SpecialChar.TokenDelimiter.value, command_line):
            (span_l, span_r) = m.span(0)
            if prev_span_r != -1:
                # New token found between prev and curr delimiter span.
                if prev_span_r <= cursor_cpos <= span_l:
                    # Found tangent token (touched by the cursor):
                    tan_token_ipos = token_ipos
                    tan_token_l_cpos = prev_span_r
                    tan_token_r_cpos = span_l

                all_tokens.append(command_line[prev_span_r:span_l])
                token_ipos += 1

            prev_span_l = span_l
            prev_span_r = span_r

        if prev_span_r == -1:
            # No delimiter spans found => entire command line is the single token.
            tan_token_ipos = token_ipos
            tan_token_l_cpos = 0
            tan_token_r_cpos = line_len
            all_tokens.append(command_line)

        tan_token = command_line[tan_token_l_cpos:tan_token_r_cpos]
        tan_token_l_part = command_line[tan_token_l_cpos:cursor_cpos]
        tan_token_r_part = command_line[cursor_cpos:tan_token_r_cpos]

        # Adjust for wrapping command line into delimiters:
        if tan_token_l_cpos != -1:
            tan_token_l_cpos -= 1
        if tan_token_r_cpos != -1:
            tan_token_r_cpos -= 1

        return (
            all_tokens,
            tan_token_ipos,
            tan_token_l_cpos,
            tan_token_r_cpos,
            tan_token,
            tan_token_l_part,
            tan_token_r_part,
        )

    def print_debug(self, end_str: str = "\n") -> None:
        if not self.is_debug_enabled:
            return
        super().print_debug("")
        eprint(TermColor.debug_output.value, end = "")
        eprint(f"tan_token_l_part: \"{self.tan_token_l_part}\"", end = " ")
        eprint(f"tan_token_r_part: \"{self.tan_token_r_part}\"", end = " ")
        eprint(f"comp_scope: {self.comp_scope.name}", end = " ")
        eprint(TermColor.reset_style.value, end = end_str)



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_data.PluginEntry import PluginEntry


@dataclass
class CheckEnvPluginConfig:
    """
    See also `CheckEnvPluginConfigSchema`.
    """

    check_env_plugin_instance_entries: dict[str, PluginEntry] = field()
    """
    Same as `plugin_instance_entries` but for FS_36_17_84_44 `check_env` only.
    """

    check_env_plugin_instance_id_activate_list: list[str] = field()
    """
    Same as `plugin_instance_id_activate_list` but for FS_36_17_84_44 `check_env` only.
    """



from argrelay.client_command_remote.ClientCommandRemoteAbstract import ClientCommandRemoteAbstract
from argrelay.enum_desc.ProcRole import ProcRole
from argrelay.relay_client.ClientCommandFactoryAbstract import (
    ClientCommandFactoryAbstract,
    select_client_response_handler,
)
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.server_spec.CallContext import CallContext


class ClientCommandFactoryRemote(ClientCommandFactoryAbstract):
    """
    Constructs communication between `ProcRole.ParentProcSpinner` and `ProcRole.ChildProcWorker`.

    Part of FS_14_59_14_06 pending requests implementation.
    """

    def __init__(
        self,
        client_config: ClientConfig,
        proc_role: ProcRole,
        w_pipe_end,
        is_optimized_completion: bool,
        server_index: int,
    ):
        self.client_config: ClientConfig = client_config
        self.proc_role: ProcRole = proc_role
        self.w_pipe_end = w_pipe_end
        self.is_optimized_completion = is_optimized_completion
        self.server_index = server_index
        if proc_role.is_worker_proc and proc_role.is_split_mode:
            assert self.w_pipe_end is not None
        else:
            assert self.w_pipe_end is None

    # noinspection PyMethodMayBeStatic
    def create_command(
        self,
        call_ctx: CallContext,
    ) -> ClientCommandRemoteAbstract:

        if self.proc_role.is_worker_proc:
            if self.is_optimized_completion:
                from argrelay.client_command_remote.ClientCommandRemoteWorkerTextProposeArgValuesOptimized import (
                    ClientCommandRemoteWorkerTextProposeArgValuesOptimized,
                )
                command_cls = ClientCommandRemoteWorkerTextProposeArgValuesOptimized
                if self.proc_role.is_split_mode:
                    from argrelay.client_pipeline.BytesSrcSender import BytesSrcSender
                    bytes_src = BytesSrcSender(
                        self.w_pipe_end,
                    )
                else:
                    from argrelay.client_pipeline.BytesSrcLocal import BytesSrcLocal
                    from argrelay.client_pipeline.BytesHandlerTextProposeArgValuesOptimized import (
                        BytesHandlerTextProposeArgValuesOptimized,
                    )
                    bytes_src = BytesSrcLocal(
                        BytesHandlerTextProposeArgValuesOptimized(),
                    )
            else:
                from argrelay.client_command_remote.ClientCommandRemoteWorkerJson import (
                    ClientCommandRemoteWorkerJson,
                )
                command_cls = ClientCommandRemoteWorkerJson
                if self.proc_role.is_split_mode:
                    from argrelay.client_pipeline.BytesSrcSender import BytesSrcSender
                    bytes_src = BytesSrcSender(
                        self.w_pipe_end,
                    )
                else:
                    from argrelay.client_pipeline.BytesSrcLocal import BytesSrcLocal
                    from argrelay.client_pipeline.BytesHandlerJson import BytesHandlerJson
                    bytes_src = BytesSrcLocal(
                        BytesHandlerJson(
                            select_client_response_handler(
                                self.proc_role,
                                call_ctx.server_action,
                            ),
                        ),
                    )
            return command_cls(
                call_ctx,
                self.proc_role,
                self.client_config.redundant_servers,
                self.server_index,
                bytes_src,
            )
        else:
            from argrelay.client_command_remote.ClientCommandRemoteSpinner import ClientCommandRemoteSpinner
            if self.is_optimized_completion:
                from argrelay.client_pipeline.BytesHandlerTextProposeArgValuesOptimized import (
                    BytesHandlerTextProposeArgValuesOptimized,
                )
                bytes_handler = BytesHandlerTextProposeArgValuesOptimized()
            else:
                from argrelay.client_pipeline.BytesHandlerJson import BytesHandlerJson
                bytes_handler = BytesHandlerJson(
                    select_client_response_handler(
                        self.proc_role,
                        call_ctx.server_action,
                    ),
                )
            return ClientCommandRemoteSpinner(
                call_ctx,
                self.proc_role,
                bytes_handler,
            )



from copy import deepcopy

from marshmallow import fields, RAISE, post_load
from marshmallow_oneofschema import OneOfSchema

from argrelay.composite_forest.CompositeNode import (
    BaseNode,
    InterpTreeNode,
    ZeroArgNode,
    FuncTreeNode,
    TreePathNode,
)
from argrelay.composite_forest.CompositeNodeType import CompositeNodeType
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FuncEnvelopeSchema import func_id_some_func_

node_type_ = "node_type"
sub_tree_ = "sub_tree"

plugin_instance_id_ = "plugin_instance_id"
func_id_ = "func_id"


class BaseNodeSchema(ObjectSchema):
    """
    Schema for (base) FS_33_76_82_84 composite forest node.
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = BaseNode

    node_type = None

    sub_tree = fields.Dict(
        keys = fields.String(),
        values = fields.Nested("CompositeNodeSchema"),
        required = False,
        allow_none = True,
        load_default = None,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        """
        Implements inheritance as described here:
        https://stackoverflow.com/a/65668854/441652
        With modification:
        *   The class of the schema instance `self` is determined by `CompositeNodeSchema.type_field` (`node_type`).
        *   The schema instance `self` contains `model_class` to create via ctor with `node_type` to pass to ctor.
        *   The process is recursive to arbitrary depth of `BaseNode.sub_tree`.
        *   Note that `BaseNodeSchema` has no `node_type` (instead, it is defined in `CompositeNodeSchema`).
        """
        return type(self).model_class(
            **input_dict,
            node_type = self.node_type,
        )


class ZeroArgNodeSchema(BaseNodeSchema):
    """
    See FS_42_76_93_51 very first zero arg mapping interp.
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = ZeroArgNode

    node_type = CompositeNodeType.zero_arg_node

    plugin_instance_id = fields.String(
        required = True,
    )


class InterpTreeNodeSchema(BaseNodeSchema):
    """
    See FS_01_89_09_24 interp tree.
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = InterpTreeNode

    node_type = CompositeNodeType.interp_tree_node

    class NextInterpSchema(ObjectSchema):
        """
        Schema internal to `InterpTreeNodeSchema`.
        """

        class Meta:
            unknown = RAISE
            strict = True

        plugin_instance_id = fields.String(
            required = True,
        )

    plugin_instance_id = fields.String(
        required = True,
    )


class FuncTreeNodeSchema(BaseNodeSchema):
    """
    See FS_26_43_73_72 func tree.
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = FuncTreeNode

    node_type = CompositeNodeType.func_tree_node

    func_id = fields.String(
        required = True,
    )


class TreePathNodeSchema(BaseNodeSchema):
    """
    See `CompositeNodeType.tree_path_node`.
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = TreePathNode

    node_type = CompositeNodeType.tree_path_node


class CompositeNodeSchema(OneOfSchema):
    """
    Polymorphic schema for composite tree node (part of composite forest) - see:
    *   https://github.com/marshmallow-code/marshmallow-oneofschema
    *   https://stackoverflow.com/q/55092906/441652
    """

    type_field = node_type_

    def get_obj_type(
        self,
        obj,
    ):
        if isinstance(obj, ZeroArgNode):
            return ZeroArgNodeSchema.node_type.name
        elif isinstance(obj, InterpTreeNode):
            return InterpTreeNodeSchema.node_type.name
        elif isinstance(obj, FuncTreeNode):
            return FuncTreeNodeSchema.node_type.name
        elif isinstance(obj, TreePathNode):
            return TreePathNodeSchema.node_type.name
        else:
            raise Exception(f"Unknown object type: {obj.__class__.__name__}")

    type_schemas = {
        ZeroArgNodeSchema.node_type.name: ZeroArgNodeSchema,
        InterpTreeNodeSchema.node_type.name: InterpTreeNodeSchema,
        FuncTreeNodeSchema.node_type.name: FuncTreeNodeSchema,
        TreePathNodeSchema.node_type.name: TreePathNodeSchema,
    }


_base_node_example = {
    node_type_: CompositeNodeType.zero_arg_node.name,
    sub_tree_: None,
}

base_node_desc = TypeDesc(
    dict_schema = BaseNodeSchema(),
    ref_name = BaseNodeSchema.__name__,
    dict_example = _base_node_example,
    default_file_path = "",
)

_zero_arg_node_example = deepcopy(base_node_desc.dict_example)
_zero_arg_node_example.update({
    plugin_instance_id_: "some_plugin_instance_id",
})

zero_arg_node_desc = TypeDesc(
    dict_schema = ZeroArgNodeSchema(),
    ref_name = ZeroArgNodeSchema.__name__,
    dict_example = _zero_arg_node_example,
    default_file_path = "",
)

_interp_tree_node_example = deepcopy(base_node_desc.dict_example)
_interp_tree_node_example.update({
    plugin_instance_id_: "some_plugin_instance_id",
})

interp_tree_node_desc = TypeDesc(
    dict_schema = InterpTreeNodeSchema(),
    ref_name = InterpTreeNodeSchema.__name__,
    dict_example = _interp_tree_node_example,
    default_file_path = "",
)

_func_tree_node_example = deepcopy(base_node_desc.dict_example)
_func_tree_node_example.update({
    func_id_: func_id_some_func_,
})

func_tree_node_desc = TypeDesc(
    dict_schema = FuncTreeNodeSchema(),
    ref_name = FuncTreeNodeSchema.__name__,
    dict_example = _func_tree_node_example,
    default_file_path = "",
)

_tree_path_node_example = deepcopy(base_node_desc.dict_example)
_tree_path_node_example.update({})

tree_path_node_desc = TypeDesc(
    dict_schema = TreePathNodeSchema(),
    ref_name = TreePathNodeSchema.__name__,
    dict_example = _tree_path_node_example,
    default_file_path = "",
)



from __future__ import annotations

from marshmallow import fields, RAISE, post_load

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.CheckEnvPluginConfig import CheckEnvPluginConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.schema_config_plugin.BasePluginConfigSchema import BasePluginConfigSchema, serialize_dag_to_list
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc

check_env_plugin_instance_entries_ = "check_env_plugin_instance_entries"


class CheckEnvPluginConfigSchema(BasePluginConfigSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = CheckEnvPluginConfig

    check_env_plugin_instance_entries = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(plugin_entry_desc.dict_schema),
        required = True,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        # Build DAG:
        check_env_plugin_instance_id_activate_order_dag = {}
        for plugin_instance_id in input_dict[check_env_plugin_instance_entries_]:
            plugin_entry: PluginEntry = input_dict[check_env_plugin_instance_entries_][plugin_instance_id]
            check_env_plugin_instance_id_activate_order_dag[plugin_instance_id] = plugin_entry.plugin_dependencies

        return type(self).model_class(
            **input_dict,
            check_env_plugin_instance_id_activate_list = serialize_dag_to_list(
                check_env_plugin_instance_id_activate_order_dag,
            ),
        )


check_env_plugin_config_desc = TypeDesc(
    dict_schema = CheckEnvPluginConfigSchema(),
    ref_name = CheckEnvPluginConfigSchema.__name__,
    dict_example = {
        check_env_plugin_instance_entries_: {
            "some_check_env_plugin_instance_id": plugin_entry_desc.dict_example,
            "some_check_env_plugin_instance_id1": plugin_entry_desc.dict_example,
            "some_check_env_plugin_instance_id2": plugin_entry_desc.dict_example,
        },
    },
    default_file_path = "check_env_plugin.conf.yaml",
)



from marshmallow import fields, RAISE

from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_response.ArgValues import ArgValues

arg_values_ = "arg_values"

_arg_values_example = {
    arg_values_: [
        "arg_value_1",
        "arg_value_2",
        "arg_value_3",
    ],
}


# TODO_11_77_28_50: Make it possible to verify proposed arg_values in all `ServerAction`-s.
# Append space to the command line (for surrogate token delimiter) in case of `ServerAction.RelayLineArgs`
# to populate proposed arg_values to allow unconditionally assert proposed values in all tests
# (for all `ServerAction`-s).

class ArgValuesSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = ArgValues

    arg_values = fields.List(
        fields.String(
            load_default = "",
        ),
        load_default = [],
        metadata = {
            "example": _arg_values_example[arg_values_],
        },
    )


arg_values_desc = TypeDesc(
    dict_schema = ArgValuesSchema(),
    ref_name = ArgValuesSchema.__name__,
    dict_example = _arg_values_example,
    default_file_path = "",
)



from typing import Union

from argrelay.handler_response.ClientResponseHandlerAbstract import ClientResponseHandlerAbstract
from argrelay.schema_response.InvocationInput import InvocationInput
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc


class ClientResponseHandlerCheckEnv(ClientResponseHandlerAbstract):
    """
    Implements FS_36_17_84_44 `check_env`.

    Unlike `ClientResponseHandlerRelayLineArgs` for regular command execution,
    this class only returns payload of the response
    (without executing client-side part of its delegator plugin).
    This is to avoid the alternative when, in order to get response data,
    stdout has to be captured from regular command execution in a sub-process.
    """

    def __init__(
        self,
    ):
        super().__init__(
        )
        self.invocation_input: Union[InvocationInput, None] = None

    def handle_response(
        self,
        response_dict: dict,
    ):
        self.invocation_input = invocation_input_desc.dict_schema.load(response_dict)



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class ConfiguratorDefaultConfig:
    """
    This class configures `ConfiguratorDefault`.
    """

    project_title: str = field(default = "")
    """
    Names an instance of `argrelay`.
    """

    project_page_url: str = field(default = "")
    """
    Project a URL to project page.
    """

    git_files_by_commit_id_url_prefix: str = field(default = "")
    """
    Provides a URL prefix to access files by commit id.

    See also `ConfiguratorAbstract.provide_project_git_files_by_commit_id_url_prefix`.
    """

    commit_id_url_prefix: str = field(default = "")
    """
    Provides a URL prefix to access page with commit id.

    See also `ConfiguratorAbstract.provide_project_commit_id_url_prefix`.
    """



import os
import signal

_child_status: int = 0
_child_exit_code: int = 0
_child_pid: int = 0
_child_exited: bool = False


def _signal_handler(signal_number, ignored_signal_frame):
    if signal_number == signal.SIGCHLD:
        global _child_status
        global _child_exit_code
        global _child_pid
        global _child_exited

        _child_exited = True
        (
            _child_pid,
            _child_status,
        ) = os.waitpid(_child_pid, os.WNOHANG)

        # Decode exit_code:
        # https://github.com/python/cpython/issues/84275#issuecomment-1093867008
        if os.WIFSIGNALED(_child_status):
            _child_exit_code = -os.WTERMSIG(_child_status)
        elif os.WIFEXITED(_child_status):
            _child_exit_code = os.WEXITSTATUS(_child_status)
        elif os.WIFSTOPPED(_child_status):
            _child_exit_code = -os.WSTOPSIG(_child_status)
        else:
            raise RuntimeError(f"unrecognized child_status: {_child_status}")


def has_child_exited() -> bool:
    global _child_exited
    return _child_exited


def is_child_successful() -> bool:
    return get_child_exit_code() == 0


def get_child_exit_code() -> int:
    global _child_exit_code
    assert has_child_exited()
    return _child_exit_code


def split_process() -> (
    bool,
    int,
    "typing.BinaryIO",
    "typing.BinaryIO",
):
    """
    Part of FS_14_59_14_06 pending requests implementation.

    The func splits parent and child processes setting up a communication pipe between the two.

    Returns tuple:
    *   bool: whether it is a parent (True) or a child (False)
    *   int: child PID for parent, None for child
    *   io: open pipe between parent and child:
        *   read-only pipe for parent
        *   write-only pipe for child

    The pipe contains data from child which, in turn, was received by child from server.

    This way, while waiting for the child to complete, the parent can do anything else (like drawing spinner).

    Both child and parent still keep their stdout and stderr connected to the terminal (this output may interleave).
    But it is expected that child prints nothing to stdout and only uses stderr in case of failure.

    NOTE: Technically, using the extra pipe between parent and child is not required
          for `ServerAction.ProposeArgValues` = Tab because Bash already captures
          stdout in that case until client completes.
          But implementation is uniform here for simplicity.
    """

    # Create pipe for the between parent and (future) child:
    r_pipe_fd, w_pipe_fd = os.pipe()

    # This seems unnecessary (probably necessary for `os.exec`), but anyway:
    # https://stackoverflow.com/q/30277081/441652
    os.set_inheritable(r_pipe_fd, True)
    os.set_inheritable(w_pipe_fd, True)

    signal.signal(signal.SIGCHLD, _signal_handler)
    child_pid: int = os.fork()

    if child_pid > 0:
        # Parent: does not write:
        os.close(w_pipe_fd)
        # Parent: reads from the pipe connected to the child:
        r_pipe_end = os.fdopen(r_pipe_fd, "rb")

        global _child_pid
        _child_pid = child_pid

        return (
            True,
            child_pid,
            r_pipe_end,
            None,
        )
    elif child_pid == 0:
        # Child: does not read:
        os.close(r_pipe_fd)
        # Child: writes into the pipe connected to the parent:
        w_pipe_end = os.fdopen(w_pipe_fd, "wb")

        return (
            False,
            0,
            None,
            w_pipe_end,
        )
    else:
        raise RuntimeError("could not fork child process")



from __future__ import annotations

import os.path
from typing import Union

from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay.client_command_remote.ClientCommandRemoteWorkerAbstract import (
    get_server_index_file_path,
    load_server_index,
    server_index_file_name,
)
from argrelay.enum_desc.ResultCategory import ResultCategory
from argrelay.enum_desc.TopDir import TopDir

_server_index_field_name = "server_index"


class PluginCheckEnvServerIndex(PluginCheckEnvAbstract):
    """
    This plugin reports `server_index` for FS_93_18_57_91 client fail over.
    """

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        at_path = f"@/{os.path.join(TopDir.var_dir.value, server_index_file_name)}"
        if os.path.isfile(get_server_index_file_path()):

            # noinspection PyBroadException
            try:
                server_index = load_server_index()
            except:
                server_index = None

            if server_index is None:
                return [CheckEnvResult(
                    result_category = ResultCategory.VerificationFailure,
                    result_key = _server_index_field_name,
                    result_value = str(None),
                    result_message = f"Remove this file as its content cannot be read as a int: {at_path}",
                )]
            else:
                return [CheckEnvResult(
                    result_category = ResultCategory.VerificationSuccess,
                    result_key = _server_index_field_name,
                    result_value = str(server_index),
                    result_message = f"See: {at_path}",
                )]
        elif os.path.exists(get_server_index_file_path()):
            return [CheckEnvResult(
                result_category = ResultCategory.VerificationFailure,
                result_key = _server_index_field_name,
                result_value = str(None),
                result_message = f"It is not a file: {at_path}",
            )]
        else:
            return [CheckEnvResult(
                result_category = ResultCategory.VerificationWarning,
                result_key = _server_index_field_name,
                result_value = str(None),
                result_message = f"The client has not been used to create this file yet: {at_path}",
            )]



from __future__ import annotations

from enum import Enum, auto

from argrelay.enum_desc.SpecialChar import SpecialChar


class TokenType(Enum):
    """
    Interpretation of tokens from command line into args (or their parts):
    *   `PosArg` = positional is specified by relative position in the command line (list of args).
    *   `KeyArg` + `ValArg` = named args are specified by pair of tokens: keyword arg token followed by value arg token.
    """

    """
    Positional arg
    """
    PosArg = auto()

    """
    Keyword arg - must have special format (e.g. ending with `:`).
    """
    KeyArg = auto()

    """
    Value arg - must be followed by `KeyArg`
    """
    ValArg = auto()


def get_token_type(all_tokens: list[str], token_ipos: int) -> TokenType:
    # TODO: POC: FS_20_88_05_60: Either remove it or implement properly: just testing named args:
    if all_tokens[token_ipos].endswith(SpecialChar.KeyValueDelimiter.value):
        return TokenType.KeyArg
    else:
        if token_ipos > 0:
            if not all_tokens[token_ipos - 1].endswith(SpecialChar.KeyValueDelimiter.value):
                return TokenType.PosArg
            else:
                return TokenType.ValArg
        else:
            return TokenType.PosArg



from __future__ import annotations

from typing import Union

from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.enum_desc.PluginType import PluginType
from argrelay.runtime_context.PluginClientAbstract import PluginClientAbstract


class PluginCheckEnvAbstract(PluginClientAbstract):
    """
    Implements FS_36_17_84_44 `check_env`.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
        )

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.LoaderPlugin

    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        """
        Executes 1 to N checks and returns results of each one.

        *   `online_mode == False` allows skipping some checks which require connection to serve.
        *   `online_mode == True` execute all checks - if connection to serve is required, such checks may fail.
        *   `online_mode is None` allows selecting default behavior (or automatic detection).
        """
        pass



from __future__ import annotations

from typing import Union

from argrelay.composite_forest.CompositeForest import CompositeForest
from argrelay.composite_forest.CompositeForestWalker import CompositeForestWalkerAbstract, TraverseDecision
from argrelay.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay.composite_forest.CompositeNode import InterpTreeNode, ZeroArgNode, TreePathNode, FuncTreeNode
from argrelay.composite_forest.DictTreeWalker import surrogate_node_id_


def extract_tree_abs_path_to_interp_id(
    composite_forest: CompositeForest,
    func_id: str,
) -> dict:
    """
    Extracts `CompositeInfoType.tree_abs_path_to_interp_id`.
    """

    tree_walker = _CompositeForestWalker_tree_abs_path_to_interp_id(
        composite_forest,
        CompositeInfoType.tree_abs_path_to_interp_id,
        func_id,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.tree_abs_path_to_interp_id


def extract_zero_arg_interp_tree(
    composite_forest: CompositeForest,
) -> dict:
    """
    Extracts `CompositeInfoType.zero_arg_interp_tree`.
    """

    tree_walker = _CompositeForestWalker_zero_arg_interp_tree(
        composite_forest,
        CompositeInfoType.zero_arg_interp_tree,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.zero_arg_interp_tree


def extract_jump_tree(
    composite_forest: CompositeForest,
) -> dict:
    """
    Extracts `CompositeInfoType.jump_tree`.
    """

    tree_walker = _CompositeForestWalker_jump_tree(
        composite_forest,
        CompositeInfoType.jump_tree,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.jump_tree


def extract_interp_tree(
    composite_forest: CompositeForest,
    plugin_instance_id: str,
) -> dict:
    """
    Extracts `CompositeInfoType.interp_tree`.
    """

    tree_walker = _CompositeForestWalker_interp_tree(
        composite_forest,
        CompositeInfoType.interp_tree,
        plugin_instance_id,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.interp_tree


def extract_func_tree(
    composite_forest: CompositeForest,
    plugin_instance_id: str,
) -> dict:
    """
    Extracts `CompositeInfoType.func_tree`.
    """

    tree_walker = _CompositeForestWalker_func_tree(
        composite_forest,
        CompositeInfoType.func_tree,
        plugin_instance_id,
    )
    tree_walker.walk_tree_roots()

    return tree_walker.func_tree


def _put_value_into_dict_path(
    given_value,
    given_path: list[str],
    target_dict: dict,
) -> None:
    curr_sub_dict: dict = target_dict

    assert len(given_path) > 0

    if len(given_path) > 1:
        for path_step in given_path[:-1]:
            if path_step not in curr_sub_dict:
                # create:
                curr_sub_dict[path_step] = {}
            curr_sub_dict = curr_sub_dict[path_step]

    curr_sub_dict[given_path[-1]] = given_value


class CompositeForestExtractorAbstract(CompositeForestWalkerAbstract):
    """
    Base class for extractor which takes specific `CompositeInfoType` from the FS_33_76_82_84 composite forest.
    """

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
    ):
        # main input:
        super().__init__(
            composite_forest,
        )

        self.info_type = info_type


# noinspection PyPep8Naming
class _CompositeForestWalker_tree_abs_path_to_interp_id(CompositeForestExtractorAbstract):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
        func_id: Union[str, None],
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # `info_type`-specific input:
        self.func_id: Union[str, None] = func_id

        # results:
        self.tree_abs_path_to_interp_id: dict = {}

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree
        # FS_91_88_07_23: jump tree:
        # Similar to `jump_path`-s, next `plugin_instance_id` configuration was drop in favour of
        # deriving such plugin automatically:
        self.next_jump_plugin_instance_id = self.curr_node.plugin_instance_id

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        if self.curr_node.func_id == self.func_id:
            assert self.curr_path[-1] == surrogate_node_id_
            assert self.curr_node.sub_tree is None
            _put_value_into_dict_path(
                self.next_jump_plugin_instance_id,
                self.curr_path[:-1],
                self.tree_abs_path_to_interp_id,
            )
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeForestWalker_zero_arg_interp_tree(CompositeForestExtractorAbstract):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # results:
        self.zero_arg_interp_tree: dict = {}

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        assert len(self.curr_path) == 1
        _put_value_into_dict_path(
            # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor:
            #       There could be multiple `zero_arg_node`-s each with its own interp,
            #       but now, there is only one interp selected by `first_interp_factory_id`.
            #       In fact, `first_interp_factory_id` must be per zero arg - see FS_15_79_76_85 line processor.
            self.curr_node.plugin_instance_id,
            self.curr_path,
            self.zero_arg_interp_tree,
        )
        self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeForestWalker_jump_tree(CompositeForestExtractorAbstract):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # results:
        self.jump_tree: dict = {}

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        assert len(self.curr_path) > 0
        # FS_91_88_07_23: jump tree:
        # So far, all `jump_path`-s are the paths leading to `curr_node` excluding that last `curr_node` step.
        # Therefore, their configuration was drop in favour of deriving `jump_path` automatically:
        jump_path = self.curr_path[:-1]
        _put_value_into_dict_path(
            jump_path,
            self.curr_path,
            self.jump_tree,
        )
        self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeForestWalker_interp_tree(CompositeForestExtractorAbstract):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
        plugin_instance_id: Union[str, None],
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # `info_type`-specific input:
        self.plugin_instance_id: Union[str, None] = plugin_instance_id

        # internal trackers:
        self.curr_parent_plugin_instance_id: Union[str, None] = None

        # results:
        self.interp_tree: dict = {}

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        if self.curr_node.plugin_instance_id == self.plugin_instance_id:
            self.curr_parent_plugin_instance_id = self.plugin_instance_id
            self.visitor_decision = TraverseDecision.walk_sub_tree
        else:
            self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        if self.curr_parent_plugin_instance_id == self.plugin_instance_id:
            _put_value_into_dict_path(
                self.curr_node.plugin_instance_id,
                self.curr_path,
                self.interp_tree,
            )
        self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self.visitor_decision = TraverseDecision.skip_sub_tree


# noinspection PyPep8Naming
class _CompositeForestWalker_func_tree(CompositeForestExtractorAbstract):

    def __init__(
        self,
        composite_forest: CompositeForest,
        info_type: CompositeInfoType,
        plugin_instance_id: Union[str, None],
    ):
        # main input:
        super().__init__(
            composite_forest,
            info_type,
        )

        # `info_type`-specific input:
        self.plugin_instance_id: Union[str, None] = plugin_instance_id

        # internal trackers:
        self.curr_parent_plugin_instance_id: Union[str, None] = None

        # results:
        self.func_tree: dict = {}

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self.visitor_decision = TraverseDecision.walk_sub_tree

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        if self.curr_node.plugin_instance_id == self.plugin_instance_id:
            self.curr_parent_plugin_instance_id = self.plugin_instance_id
            self.visitor_decision = TraverseDecision.walk_sub_tree
        else:
            self.visitor_decision = TraverseDecision.skip_sub_tree

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        if self.curr_parent_plugin_instance_id == self.plugin_instance_id:
            _put_value_into_dict_path(
                self.curr_node.func_id,
                self.curr_path,
                self.func_tree,
            )
        self.visitor_decision = TraverseDecision.skip_sub_tree



# Keep minimal import on start:
import sys

from argrelay.client_spec.ShellContext import ShellContext
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.ProcRole import ProcRole
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.misc_helper_common import get_config_path
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_client.client_utils import handle_main_exception
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig

ElapsedTime.measure("after_program_entry")


def main():
    # noinspection PyBroadException
    try:
        return run_client()
    except BaseException as e1:
        handle_main_exception(e1)


def run_client():
    ElapsedTime.measure("after_initial_imports")

    file_path = get_config_path("argrelay_client.json")
    client_config = load_client_config(file_path)
    ElapsedTime.measure("after_loading_client_config")

    shell_ctx = ShellContext.from_env(sys.argv)
    shell_ctx.print_debug()
    call_ctx = shell_ctx.create_call_context()

    is_split_mode: bool = (
        # FS_14_59_14_06 pending requests: at the moment the process is split only to provide spinner:
        client_config.show_pending_spinner
        and
        shell_ctx.comp_type is not CompType.InvokeAction
    )

    is_optimized_completion = (
        call_ctx.server_action is ServerAction.ProposeArgValues
        and
        client_config.optimize_completion_request
    )

    w_pipe_end = None
    if is_split_mode:
        from argrelay.relay_client.proc_splitter import split_process
        (
            is_parent,
            child_pid,
            r_pipe_end,
            w_pipe_end,
        ) = split_process()

        if is_parent:
            proc_role: ProcRole = ProcRole.ParentProcSpinner
            from argrelay.relay_client.proc_spinner import spinner_main
            spinner_main(
                call_ctx,
                client_config,
                proc_role,
                is_optimized_completion,
                r_pipe_end,
            )
            return None
        else:
            proc_role: ProcRole = ProcRole.ChildProcWorker
    else:
        proc_role: ProcRole = ProcRole.SoleProcWorker

    from argrelay.relay_client.proc_worker import worker_main
    command_obj = worker_main(
        call_ctx,
        client_config,
        proc_role,
        is_optimized_completion,
        w_pipe_end,
        shell_ctx,
        -1,
    )
    return command_obj


def load_client_config(file_path):
    import json
    with open(file_path) as config_file:
        client_config_dict = json.load(config_file)
    client_config = client_config_dict_to_object(client_config_dict)
    return client_config


def client_config_dict_to_object(client_config_dict):
    """
    Optimized dict -> object conversion avoiding import of `*Schema` for performance:
    """
    redundant_servers: list[ConnectionConfig] = []
    for redundant_server in client_config_dict["redundant_servers"]:
        redundant_servers.append(ConnectionConfig(
            server_host_name = redundant_server["server_host_name"],
            server_port_number = redundant_server["server_port_number"],
        ))
    client_config = ClientConfig(
        __comment__ = client_config_dict.get("use_local_requests", None),
        use_local_requests = client_config_dict.get("use_local_requests", False),
        optimize_completion_request = client_config_dict.get("optimize_completion_request", True),
        redundant_servers = redundant_servers,
        show_pending_spinner = client_config_dict.get("show_pending_spinner", True),
        spinless_sleep_sec = client_config_dict.get("spinless_sleep_sec", 0.0),
    )
    return client_config


if __name__ == "__main__":
    main()



import time
from datetime import datetime

from dateutil.tz import tz

from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvServerResponseValueAbstract import PluginCheckEnvServerResponseValueAbstract
from argrelay.check_env.check_env_utils import format_time_to_relative
from argrelay.custom_integ.SchemaPluginCheckEvnServerResponseValueAbstract import field_values_to_command_lines_
from argrelay.enum_desc.CheckEnvField import CheckEnvField


class PluginCheckEnvServerResponseValueStartTime(PluginCheckEnvServerResponseValueAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict = plugin_config_dict or {
                field_values_to_command_lines_: {
                    CheckEnvField.server_start_time.name: "argrelay.check_env server_start_time",
                },
            },
        )

    def verify_online_value(
        self,
        field_name,
        field_value,
    ) -> CheckEnvResult:
        check_env_result: CheckEnvResult = super().verify_online_value(
            field_name,
            field_value,
        )
        # Put formated time to the message:
        if check_env_result.result_category.VerificationSuccess:
            from_zone = tz.tzutc()
            into_zone = tz.tzlocal()
            utc_time = datetime.utcfromtimestamp(int(field_value))
            utc_time = utc_time.replace(tzinfo = from_zone)

            epoch_start = datetime.utcfromtimestamp(0)

            abs_time_str = utc_time.astimezone(into_zone).isoformat()
            rel_time_str = format_time_to_relative(
                time.time() * 1000,
                (utc_time.replace(tzinfo = None) - epoch_start).total_seconds() * 1000,
            )

            check_env_result.result_message = f"{abs_time_str} ~ {rel_time_str}"
        return check_env_result



from enum import Enum, auto


class SpecialFunc(Enum):
    """
    See also: FS_80_45_89_81 / integrated functions
    """

    func_id_unplugged = auto()
    """
    Function which is not plugged anywhere.
    """

    func_id_no_data = auto()
    """
    Function which has no data it searches for via FS_31_70_49_15 `search_control`.
    """

    func_id_intercept_invocation = auto()
    """
    Implements FS_88_66_66_73 intercept func.
    """

    func_id_help_hint = auto()
    """
    Implements FS_71_87_33_52 help hint.
    """

    func_id_echo_args = auto()
    """
    Implements FS_43_50_57_71 `echo_args` func.
    """

    func_id_query_enum_items = auto()
    """
    Implements FS_02_25_41_81 `func_id_query_enum_items`.
    """



import sys


def handle_main_exception(e1):
    # noinspection PyBroadException
    try:
        # Avoid leaving terminal in unexpected state.
        # For example, due to some terminal control chars printed by client partially,
        # the terminal may be left in mode which does not echo back chars typed by the user.
        if sys.stdout.isatty() or sys.stderr.isatty():
            import os
            os.system("stty sane")
    except BaseException as e2:
        pass
    raise e1



from __future__ import annotations

import copy
import os
from datetime import timezone, datetime, timedelta

from git import Repo

from argrelay.custom_integ.GitRepoDelegator import repo_root_abs_path_
from argrelay.custom_integ.GitRepoEntryConfigSchema import (
    repo_rel_path_,
    envelope_properties_,
    is_repo_enabled_,
    load_repo_commits_,
    load_repo_tags_,
    load_commits_max_count_,
    load_tags_last_days_,
)
from argrelay.custom_integ.GitRepoEnvelopeClass import GitRepoEnvelopeClass
from argrelay.custom_integ.GitRepoLoaderConfigSchema import (
    git_repo_loader_config_desc,
    load_git_commits_default_,
    repo_entries_,
    load_git_tags_default_,
)
from argrelay.custom_integ.GitRepoPropName import GitRepoPropName
from argrelay.custom_integ.git_utils import is_git_repo
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.misc_helper_common import eprint, get_argrelay_dir
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_id_,
    envelope_payload_,
)


def normalize_single_line_to_black_spaces(
    s: str,
) -> str:
    """
    Get rid of white spaces.

    See FS_99_81_19_25 no space in options.
    """
    return s.strip().replace(" ", "_")


def normalize_multiple_lines_to_black_spaces(
    s: str,
) -> str:
    """
    Get rid of multiple lines (take the first one only) and white spaces.

    See FS_99_81_19_25 no space in options.
    """
    return s.partition('\n')[0].strip().replace(" ", "_")


class GitRepoLoader(AbstractLoader):
    """
    Implements FS_67_16_61_97 git_plugin.
    """

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return git_repo_loader_config_desc.dict_from_input_dict(plugin_config_dict)

    def update_static_data(
        self,
        static_data: StaticData,
        query_engine: QueryEngine,
    ) -> StaticData:

        if not self.plugin_config_dict:
            return static_data

        static_data = self.load_git_objects(static_data)

        return static_data

    def load_git_objects(
        self,
        static_data: StaticData,
    ) -> StaticData:
        """
        Scan `base_path` recursively and load metadata about all Git repos found.
        """

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        class_names = [
            GitRepoEnvelopeClass.ClassGitRepo.name,
            GitRepoEnvelopeClass.ClassGitTag.name,
            GitRepoEnvelopeClass.ClassGitCommit.name,
        ]

        init_envelop_collections(
            self.server_config,
            class_names,
            # Same index fields for all collections (can be fine-tuned later):
            lambda collection_name, class_name: [enum_item.name for enum_item in GitRepoPropName]
        )

        repo_envelopes = static_data.envelope_collections[
            class_to_collection_map[GitRepoEnvelopeClass.ClassGitRepo.name]
        ].data_envelopes
        tag_envelopes = static_data.envelope_collections[
            class_to_collection_map[GitRepoEnvelopeClass.ClassGitTag.name]
        ].data_envelopes
        commit_envelopes = static_data.envelope_collections[
            class_to_collection_map[GitRepoEnvelopeClass.ClassGitCommit.name]
        ].data_envelopes

        load_git_commits_default = self.plugin_config_dict[load_git_commits_default_]
        load_git_tags_default = self.plugin_config_dict[load_git_tags_default_]

        # List of registered Git abs paths:
        repo_root_abs_paths = []
        for repo_base_path in self.plugin_config_dict[repo_entries_]:

            repo_entries = self.plugin_config_dict[repo_entries_][repo_base_path]
            repo_base_path = os.path.expanduser(repo_base_path)
            if os.path.isabs(repo_base_path):
                repo_base_abs_path = repo_base_path
            else:
                repo_base_abs_path = os.path.join(get_argrelay_dir(), repo_base_path)

            # Process repo entries collecting `repo_root_rel_path` and `repo_root_abs_path`:
            for repo_entry in repo_entries:

                repo_root_rel_path: str = repo_entry[repo_rel_path_] or "."
                repo_root_abs_path: str = os.path.join(repo_base_abs_path, repo_root_rel_path)

                if not repo_entry[is_repo_enabled_]:
                    eprint(f"INFO: disabled repo: {repo_root_abs_path}")
                    continue

                # Deduplicate Git root paths:
                if repo_root_abs_path not in repo_root_abs_paths:
                    # Register `repo_root_abs_path`:
                    repo_root_abs_paths.append(repo_root_abs_path)
                else:
                    eprint(f"ERROR: duplicate Git repo: {repo_root_abs_path}")
                    raise RuntimeError

                # Query Git root path of curr dir:
                if is_git_repo(
                    repo_root_abs_path,
                ):
                    pass
                else:
                    eprint(f"ERROR: not a Git repo: {repo_root_abs_path}")
                    # Walked into dir outside of Git repo:
                    continue

                repo_root_base_name = os.path.basename(repo_root_abs_path)

                ############################################################################################################
                # repos

                repo_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                repo_envelope.update({
                    envelope_id_: f"{repo_root_abs_path}:{GitRepoEnvelopeClass.ClassGitRepo.name}",
                    envelope_payload_: {
                        repo_root_abs_path_: repo_root_abs_path,
                    },
                    ReservedPropName.envelope_class.name: GitRepoEnvelopeClass.ClassGitRepo.name,
                    GitRepoPropName.git_repo_root_rel_path.name: repo_root_rel_path,
                    GitRepoPropName.git_repo_root_abs_path.name: repo_root_abs_path,
                    GitRepoPropName.git_repo_root_base_name.name: repo_root_base_name,
                })
                self.enrich_git_repo_object(repo_envelope)
                repo_envelopes.append(repo_envelope)

                load_repo_tags = repo_entry[load_repo_tags_]
                if load_git_tags_default or load_repo_tags:

                    ########################################################################################################
                    # tags

                    load_tags_last_days = repo_entry[load_tags_last_days_]

                    earliest_ts_utc = (datetime.now() - timedelta(days = load_tags_last_days)).astimezone(timezone.utc)

                    git_repo = Repo(repo_root_abs_path)

                    for git_tag in git_repo.tags:

                        git_commit = git_tag.commit

                        (
                            commit_timestamp_utc,
                            commit_date_utc,
                            commit_time_utc,
                        ) = self.get_commit_date_and_time(git_commit)

                        if commit_timestamp_utc < earliest_ts_utc:
                            continue

                        tag_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                        tag_envelope.update({
                            envelope_id_: f"{repo_root_abs_path}:{GitRepoEnvelopeClass.ClassGitTag.name}:{git_tag.name}",
                            envelope_payload_: {
                            },
                            ReservedPropName.envelope_class.name: GitRepoEnvelopeClass.ClassGitTag.name,
                            GitRepoPropName.git_repo_root_rel_path.name: repo_root_rel_path,
                            GitRepoPropName.git_repo_root_abs_path.name: repo_root_abs_path,
                            GitRepoPropName.git_repo_root_base_name.name: repo_root_base_name,
                            #
                            GitRepoPropName.git_repo_tag_name.name: git_tag.name,
                            #
                            GitRepoPropName.git_repo_commit_id.name: git_commit.hexsha,
                            GitRepoPropName.git_repo_short_commit_id.name: git_commit.hexsha[:7],
                            GitRepoPropName.git_repo_commit_author_name.name: normalize_single_line_to_black_spaces(
                                git_commit.author.name,
                            ),
                            GitRepoPropName.git_repo_commit_author_email.name: git_commit.author.email.lower(),
                            GitRepoPropName.git_repo_commit_message.name: normalize_multiple_lines_to_black_spaces(
                                git_commit.message,
                            ),
                            GitRepoPropName.git_repo_commit_date.name: commit_date_utc,
                            GitRepoPropName.git_repo_commit_time.name: commit_time_utc,
                        })
                        self.enrich_git_repo_object(tag_envelope)
                        tag_envelopes.append(tag_envelope)

                load_repo_commits = repo_entry[load_repo_commits_]
                if load_git_commits_default or load_repo_commits:

                    ########################################################################################################
                    # commits

                    load_commits_max_count = repo_entry[load_commits_max_count_]

                    git_repo = Repo(repo_root_abs_path)
                    for git_commit in git_repo.iter_commits(max_count = load_commits_max_count):
                        (
                            commit_timestamp_utc,
                            commit_date_utc,
                            commit_time_utc,
                        ) = self.get_commit_date_and_time(git_commit)

                        commit_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                        commit_envelope.update({
                            envelope_id_: f"{repo_root_abs_path}:{GitRepoEnvelopeClass.ClassGitCommit.name}:{git_commit.hexsha}",
                            envelope_payload_: {
                            },
                            ReservedPropName.envelope_class.name: GitRepoEnvelopeClass.ClassGitCommit.name,
                            GitRepoPropName.git_repo_root_rel_path.name: repo_root_rel_path,
                            GitRepoPropName.git_repo_root_abs_path.name: repo_root_abs_path,
                            GitRepoPropName.git_repo_root_base_name.name: repo_root_base_name,
                            #
                            GitRepoPropName.git_repo_commit_id.name: git_commit.hexsha,
                            GitRepoPropName.git_repo_short_commit_id.name: git_commit.hexsha[:7],
                            GitRepoPropName.git_repo_commit_author_name.name: normalize_single_line_to_black_spaces(
                                git_commit.author.name,
                            ),
                            GitRepoPropName.git_repo_commit_author_email.name: git_commit.author.email,
                            GitRepoPropName.git_repo_commit_message.name: normalize_multiple_lines_to_black_spaces(
                                git_commit.message,
                            ),
                            GitRepoPropName.git_repo_commit_date.name: commit_date_utc,
                            GitRepoPropName.git_repo_commit_time.name: commit_time_utc,
                        })
                        self.enrich_git_repo_object(commit_envelope)
                        commit_envelopes.append(commit_envelope)

        return static_data

    # noinspection PyMethodMayBeStatic
    def get_commit_date_and_time(
        self,
        git_commit,
    ):
        commit_timestamp_utc = git_commit.committed_datetime.astimezone(timezone.utc)
        commit_date_utc = f"{commit_timestamp_utc.date().isoformat()}Z"
        commit_time_utc = f"{commit_timestamp_utc.time().isoformat()}Z"
        return (
            commit_timestamp_utc,
            commit_date_utc,
            commit_time_utc,
        )

    def enrich_git_repo_object(
        self,
        data_envelope: dict,
    ) -> None:
        object_categories: list[str] = self.categorize_git_object(data_envelope)

        if not object_categories:
            object_categories.append("UNKNOWN_category")

        data_envelope.update({
            GitRepoPropName.git_repo_object_category.name: object_categories,
        })

    def categorize_git_object(
        self,
        data_envelope: dict,
    ) -> list[str]:
        object_class = data_envelope[ReservedPropName.envelope_class.name]
        if object_class == GitRepoEnvelopeClass.ClassGitTag.name:
            return self.categorize_git_tag(data_envelope)
        return []

    def categorize_git_tag(
        self,
        data_envelope: dict,
    ) -> list[str]:
        tag_categories = []
        tag_name = data_envelope[GitRepoPropName.git_repo_tag_name.name]

        if ".dev" in tag_name:
            tag_categories.append("pre-release")
        if tag_name.endswith(".final"):
            tag_categories.append("release")

        if tag_name.startswith("v0."):
            tag_categories.append("unstable")
        else:
            tag_categories.append("stable")

        return tag_categories



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class QueryResult:
    """
    Results of a query via `QueryEngine.query_prop_values`.

    Do not confuse with results of `QueryEngine.query_data_envelopes` (which returns all `data_envelope`-s directly).
    """

    data_envelopes: list[dict] = field()
    """
    Contains 0 or 1 `data_envelope`-s from search results.

    The contents of the `data_envelopes` is only reliable (the `data_envelope` found) when `found_count == 1`.

    Returning maximum single `data_envelope` (when `found_count` != 0) is used for
    performance reasons with `QueryEngine.query_prop_values`.

    See also `EnvelopeContainer.data_envelopes` to get the entire list of the `data_envelopes`.
    """

    found_count: int = field()
    """
    Total number of `data_envelope`-s found - see `data_envelopes`.

    When `found_count == 1`, `data_envelopes` contains exactly the one `data_envelope` found.
    """

    remaining_types_to_values: dict[str, list[str]] = field()
    """
    See `EnvelopeContainer.remaining_types_to_values`.

    Each list is sorted.
    """



from __future__ import annotations

from argrelay.runtime_context.AbstractPlugin import AbstractPlugin, import_plugin_class
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.ServerConfig import ServerConfig


class AbstractPluginServer(AbstractPlugin):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        self.server_config: ServerConfig = server_config
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
        )

    def provide_plugin_search_control(
        self,
    ) -> list[SearchControl]:
        """
        Report `search_control`, if plugin uses any.

        If plugin uses `search_control`, it is supposed to report it so that server can validate `data_envelope`-s.
        """
        return []

    def validate_loaded_data(
        self,
        static_data: "StaticData"
    ) -> None:
        """
        Callback to validate data after all :class:`AbstractLoader`-s completed loading data.

        This is a chance for any plugin to verify that the data still matches its expectation.
        """
        pass


def instantiate_server_plugin(
    server_config: ServerConfig,
    plugin_instance_id: str,
    plugin_entry,
):
    plugin_class = import_plugin_class(plugin_entry)
    plugin_object: AbstractPlugin = plugin_class(
        server_config,
        plugin_instance_id,
        plugin_entry.plugin_config,
    )
    return plugin_object



from __future__ import annotations

import contextlib
from copy import deepcopy
from typing import Union

from argrelay.composite_forest.CompositeForestExtractor import extract_interp_tree
from argrelay.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay.composite_forest.DictTreeWalker import DictTreeWalker, sequence_starts_with, fetch_subtree_node, \
    surrogate_node_id_
from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.PluginType import PluginType
from argrelay.misc_helper_server import insert_unique_to_sorted_list
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory, AbstractInterp
from argrelay.plugin_interp.InterpTreeInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
)
from argrelay.plugin_interp.NoopInterpFactory import NoopInterpFactory
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig, assert_plugin_instance_id


class InterpTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.is_recursive_func_load = False
        self.interp_selector_tree = extract_interp_tree(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_instance_id,
        )

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return tree_path_interp_factory_config_desc.dict_from_input_dict(plugin_config_dict)

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor:
        #       At the moment, `InterpTreeInterpFactory` is not plugged into main interp tree -
        #       instead, it is plugged by `FirstArgInterpFactory`.

        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.interp_tree,
            self.interp_selector_tree,
        )
        # Walk configured interp tree and call `load_interp_tree_abs_paths` with `interp_tree_abs_path` for each interp.
        all_interp_tree_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()
        for interp_factory_instance_id in all_interp_tree_abs_paths:
            assert_plugin_instance_id(
                self.server_config,
                interp_factory_instance_id,
                PluginType.InterpFactoryPlugin,
            )
            interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_factory_instance_id]
            interp_factory.load_interp_tree_abs_paths([
                tuple(list_abs_path)
                for list_abs_path in all_interp_tree_abs_paths[interp_factory_instance_id]
            ])

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        with self._is_recursive_load() as is_recursive_load:
            # TODO: FS_33_76_82_84 composite forest: add validation that same interp tree interp is not loaded twice - is it required (given that plugin instances can be reused)?
            if is_recursive_load:
                return []
            return self._load_func_envelopes(
                interp_tree_abs_path,
                func_ids_to_func_envelopes,
            )

    @contextlib.contextmanager
    def _is_recursive_load(self):
        """
        Auto-reset flag `self.is_recursive_func_load` on exit from `with` section.
        """
        try:
            if not self.is_recursive_func_load:
                self.is_recursive_func_load = True
                is_recursive_load = False
            else:
                is_recursive_load = True
            yield is_recursive_load
        finally:
            self.is_recursive_func_load = False

    def _load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        mapped_func_ids: list[str] = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.interp_tree,
            self.interp_selector_tree,
        )
        # Walk configured interp tree and call `load_func_envelopes` with `interp_tree_abs_path` for each interp.
        all_interp_tree_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()
        for interp_factory_instance_id in all_interp_tree_abs_paths:
            for sub_interp_tree_abs_path in all_interp_tree_abs_paths[interp_factory_instance_id]:
                if not sequence_starts_with(sub_interp_tree_abs_path, interp_tree_abs_path):
                    # skip: other `interp_tree_abs_path`-s are going to be separate calls to this func:
                    continue
                assert_plugin_instance_id(
                    self.server_config,
                    interp_factory_instance_id,
                    PluginType.InterpFactoryPlugin,
                )
                interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_factory_instance_id]

                mapped_func_ids.extend(interp_factory.load_func_envelopes(
                    tuple(sub_interp_tree_abs_path),
                    func_ids_to_func_envelopes,
                ))
        return mapped_func_ids

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> InterpTreeInterp:
        """
        Each `InterpTreeInterpFactory` serves a subtree,
        but it only stores config of per its plug points
        (interp tree abs paths where it was invoked on `load_func_envelopes`).
        If next interp with its plugin id is created, but the specified path is within its subtree,
        chop off the last step in the path until it matches path to its subtree to select one.
        For example:
        ```
        # Current path within interp tree:
        interp_ctx.interp_tree_abs_path = ('l1', 'l2', 'l3', 'l4', )
        # Plug point of this factory within the interp tree:
        self.interp_tree_abs_paths_to_node_configs = {
            ('l1', 'l2', ): {},
            ('x1', 'x2', 'x3', ): {},
        }
        # Keep trying to cut of the last step in the path until it matches one of the key:
        ('l1', 'l2', 'l3', 'l4', )
        ('l1', 'l2', 'l3', )
        ('l1', 'l2', )
        # Done.
        ```
        """
        orig_subtree_abs_path = interp_ctx.interp_tree_abs_path
        curr_subtree_abs_path = orig_subtree_abs_path
        while curr_subtree_abs_path not in self.interp_tree_abs_paths_to_node_configs:
            if len(curr_subtree_abs_path) == 0:
                raise ValueError(
                    f"path `{orig_subtree_abs_path}` is not subpath of `{self.interp_tree_abs_paths_to_node_configs.keys()}` paths"
                )
            curr_subtree_abs_path = curr_subtree_abs_path[:-1]
        return InterpTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[curr_subtree_abs_path],
            interp_ctx,
            self.interp_selector_tree,
        )


class InterpTreeInterp(AbstractInterp):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        interp_selector_tree: dict,
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
        )
        self.interp_selector_tree: dict = interp_selector_tree
        self.next_interp_factory_id: Union[str, None] = None

        # TODO: Why hard-coded? Isn't it possible for this plugin to be plugged into any depth of the tree?
        # Token with ipos = 0 is the command name eaten by `FirstArgInterp` (FS_42_76_93_51 first interp):
        self.base_token_ipos: int = 1

    def consumes_args_at_once(self) -> bool:
        return True

    def consume_pos_args(self) -> bool:
        """
        Consumes heading args in `remaining_arg_buckets` according to the `interp_selector_tree`.

        Unlike normal guideline to consume one arg at time (FS_44_36_84_88),
        this func consumes all possible args
        because args are selected according to the `interp_selector_tree` (not via query)
        and do not become incompatible by consuming all (causing FS_51_67_38_37 impossible arg combinations).

        Also, FS_01_89_09_24 interp tree does not observe FS_97_64_39_94 `arg_bucket` boundaries
        and consumes all necessary args sequentially.
        """

        curr_sub_tree = fetch_subtree_node(
            self.interp_selector_tree,
            self.interp_tree_abs_path,
        )
        any_consumed = False
        while True:
            if isinstance(curr_sub_tree, str):
                # Tree leaf is reached - use it:
                self.next_interp_factory_id = curr_sub_tree
                return any_consumed

            # Always consume next remaining token:
            # TODO: Is this assumption valid/safe that next remaining `ipos` is in the order it appears on command line?
            #       Apparently, it is fine as we keep deleting head of `remaining_arg_buckets` below:
            curr_token_ipos = self.interp_ctx.next_remaining_token_ipos()

            if curr_token_ipos is None:
                self.set_default_factory_id(curr_sub_tree)
                return any_consumed

            curr_token_value = self.interp_ctx.parsed_ctx.all_tokens[curr_token_ipos]
            assert self.is_pos_arg(curr_token_ipos)

            if isinstance(curr_sub_tree, dict):
                if curr_token_value in curr_sub_tree:
                    # Consume one more arg into path:
                    self.interp_tree_abs_path.append(curr_token_value)

                    bucket_index = self.interp_ctx.token_ipos_to_arg_bucket_map[curr_token_ipos]
                    self.interp_ctx.consumed_arg_buckets[bucket_index].append(curr_token_ipos)
                    del self.interp_ctx.remaining_arg_buckets[bucket_index][0]

                    curr_sub_tree = curr_sub_tree[curr_token_value]
                    any_consumed = True
                    continue
                else:
                    self.set_default_factory_id(curr_sub_tree)
                    return any_consumed
            else:
                raise LookupError()

    def set_default_factory_id(
        self,
        curr_sub_tree,
    ):
        # Impossible to consume more arg - use default of the current sub-tree:
        if surrogate_node_id_ in curr_sub_tree:
            self.next_interp_factory_id = curr_sub_tree[surrogate_node_id_]
        else:
            # TODO: TODO_62_75_33_41: Do not hardcode plugin instance id (instance of `NoopInterpFactory`):
            self.next_interp_factory_id = f"{NoopInterpFactory.__name__}.default"

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp

    def next_interp(self) -> "AbstractInterp":
        # Selected `next_interp_factory_id` need to be given one of the paths in the FS_01_89_09_24 interp tree
        # (because each `next_interp_factory_id` can be plugged into multiple interp tree leaves).
        self.interp_ctx.interp_tree_abs_path = tuple(self.interp_tree_abs_path)
        return self.interp_ctx.create_next_interp(self.next_interp_factory_id)

    def propose_arg_completion(self) -> None:
        """
        Suggest based on current sub-tree from config.
        """

        if not self.is_eligible_for_suggestion():
            return

        curr_sub_tree = fetch_subtree_node(
            self.interp_selector_tree,
            self.interp_tree_abs_path,
        )
        if isinstance(curr_sub_tree, dict):
            for proposed_value in [
                x for x in curr_sub_tree
                if (
                    isinstance(x, str)
                    and
                    x != surrogate_node_id_
                    and
                    # FS_32_05_46_00: using `startswith`:
                    x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                )
            ]:
                insert_unique_to_sorted_list(self.interp_ctx.comp_suggestions, proposed_value)

    def is_eligible_for_suggestion(self):
        """
        Suggesting anything is possible only if there is no other tokens
        (after those recorded in `interp_tree_abs_path`) available for consumption by subsequent interpreters.
        """

        if self.interp_ctx.next_remaining_token_ipos() is None:
            return True
        else:
            return False

        # TODO: Clean up code below or take into account tangent token (ipos cursor position).
        #       Note that there is no ipos cursor position at the moment.
        #       It has to be computed with help of FS_23_62_89_43 tangent token
        #       (which might be a surrogate one if cursor does not touch non-whitespace chars).
        remaining_consumed_tokens = deepcopy(self.interp_ctx.consumed_token_ipos_list())

        # Remove everything until `base_token_ipos`:
        for token_ipos in range(0, self.base_token_ipos):
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)

        # Remove everything eaten into `interp_tree_abs_path`:
        token_ipos = self.base_token_ipos
        for node_path_token in self.interp_tree_abs_path:
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)
            token_ipos += 1

        # Is there anything else?
        return len(remaining_consumed_tokens) == 0



from __future__ import annotations

from typing import Union

from argrelay.composite_forest.CompositeForestExtractor import extract_tree_abs_path_to_interp_id
from argrelay.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay.composite_forest.DictTreeWalker import DictTreeWalker
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_delegator.AbstractJumpDelegatorConfigSchema import (
    abstract_jump_delegator_config_desc,
    single_func_id_,
)
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterp
from argrelay.runtime_data.ServerConfig import ServerConfig


class AbstractJumpDelegator(AbstractDelegator):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        extracted_dict = extract_tree_abs_path_to_interp_id(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_config_dict[single_func_id_],
        )
        dict_tree_walker = DictTreeWalker(
            CompositeInfoType.tree_abs_path_to_interp_id,
            extracted_dict,
        )
        # Temporary (reversed) map which contains path per id (instead of id per path):
        temporary_id_to_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()

        # Reverse temporary path per id map into id per path map:
        self.tree_path_to_next_interp_plugin_instance_id: dict[tuple[str, ...], str] = {}
        for interp_factory_instance_id, tree_abs_paths in temporary_id_to_paths.items():
            for tree_abs_path in tree_abs_paths:
                self.tree_path_to_next_interp_plugin_instance_id[tuple(tree_abs_path)] = interp_factory_instance_id

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return abstract_jump_delegator_config_desc.dict_from_input_dict(plugin_config_dict)

    def run_interp_control(
        self,
        curr_interp: AbstractInterp,
    ) -> Union[None, str]:
        # TODO_10_72_28_05: support special funcs for all commands:
        #                   Delegator must select next `interp_factory_id` based on `interp_tree_abs_path` via `tree_abs_path_to_interp_id` (not based on single plugin id specified)
        #                   because delegator can be accessible through multiple tree paths.
        #                   Delegator should map specific tree path using the tree path they are accessed through to specific `interp_plugin_instance_id`.
        #                   The next interp selection is a double jump:
        #                   *   the first jump here (in delegator) is based on selected func via its delegator to interp,
        #                   *   the second jump there (in jump tree interp) from jump interp via jump tree.
        # TODO: This must be special interpreter which is configured only to search functions (without their args).
        #       NEXT TODO: Why not support function args for special interp (e.g. `intercept` or `help` with format params)?
        #                  The interp_control is only run when func and all its args are selected and there should be next interp to continue.
        if curr_interp.interp_ctx.interp_tree_abs_path in self.tree_path_to_next_interp_plugin_instance_id:
            return self.tree_path_to_next_interp_plugin_instance_id[curr_interp.interp_ctx.interp_tree_abs_path]
        else:
            return None



"""
"""

func_id_goto_host_ = "func_id_goto_host"
func_id_goto_service_ = "func_id_goto_service"
func_id_goto_git_repo_ = "func_id_goto_git_repo"

func_id_desc_host_ = "func_id_desc_host"
func_id_desc_service_ = "func_id_desc_service"

func_id_list_host_ = "func_id_list_host"
func_id_list_service_ = "func_id_list_service"

func_id_diff_service_ = "func_id_diff_service"

func_id_desc_git_tag_ = "func_id_desc_git_tag"
func_id_desc_git_commit_ = "func_id_desc_git_commit"



from __future__ import annotations

from marshmallow import fields, RAISE

from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_client.ConnectionConfigSchema import connection_config_desc
from argrelay.schema_config_core_server.GuiBannerConfigSchema import gui_banner_config_desc
from argrelay.schema_config_core_server.MongoConfigSchema import mongo_config_desc
from argrelay.schema_config_core_server.QueryCacheConfigSchema import query_cache_config_desc
from argrelay.schema_config_core_server.ServerPluginControlSchema import server_plugin_control_desc
from argrelay.schema_config_core_server.StaticDataSchema import static_data_desc

connection_config_ = "connection_config"
mongo_config_ = "mongo_config"
query_cache_config_ = "query_cache_config"
gui_banner_config_ = "gui_banner_config"
default_gui_command_ = "default_gui_command"
class_to_collection_map_ = "class_to_collection_map"
server_plugin_control_ = "server_plugin_control"
static_data_ = "static_data"


class ServerConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = ServerConfig

    connection_config = fields.Nested(
        connection_config_desc.dict_schema,
        required = True,
    )

    mongo_config = fields.Nested(
        mongo_config_desc.dict_schema,
        required = True,
    )

    query_cache_config = fields.Nested(
        query_cache_config_desc.dict_schema,
        required = True,
    )

    gui_banner_config = fields.Nested(
        gui_banner_config_desc.dict_schema,
        required = True,
    )

    default_gui_command = fields.String(
        required = False,
        load_default = None,
    )

    # Related to FS_56_43_05_79 search diff collection:
    # Maps names of envelope classes to names of `EnvelopeCollection`.
    # Specifically, it allows to:
    # *   map classes to put all into single collection
    # *   map classes to put each into separate collection
    # Normally, if map entry is missing for one of the envelope class,
    # it uses envelope class name as collection name by default.
    class_to_collection_map = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )

    server_plugin_control = fields.Nested(
        server_plugin_control_desc.dict_schema,
        required = True,
    )

    # TODO_00_79_72_55: remove in the future:
    static_data = fields.Nested(
        static_data_desc.dict_schema,
        required = False,
        load_default = StaticData(
            envelope_collections = {},
        ),
    )


server_config_desc = TypeDesc(
    dict_schema = ServerConfigSchema(),
    ref_name = ServerConfigSchema.__name__,
    dict_example = {
        connection_config_: connection_config_desc.dict_example,
        mongo_config_: mongo_config_desc.dict_example,
        query_cache_config_: query_cache_config_desc.dict_example,
        gui_banner_config_: gui_banner_config_desc.dict_example,
        class_to_collection_map_: {
            ServiceEnvelopeClass.ClassCluster.name: ServiceEnvelopeClass.ClassCluster.name,
            ServiceEnvelopeClass.ClassHost.name: ServiceEnvelopeClass.ClassHost.name,
            ServiceEnvelopeClass.ClassService.name: ServiceEnvelopeClass.ClassService.name,
            ServiceEnvelopeClass.ClassAccessType.name: ServiceEnvelopeClass.ClassAccessType.name,
        },
        server_plugin_control_: server_plugin_control_desc.dict_example,
        static_data_: static_data_desc.dict_example,
    },
    default_file_path = "argrelay_server.yaml",
)



from argrelay.relay_client.ClientAbstract import ClientAbstract
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.test_infra.ClientCommandFactoryLocal import ClientCommandFactoryLocal


class ClientLocal(ClientAbstract):
    """
    Client local to server

    `ClientLocal` always talks to `LocalServer` via direct func calls (without communication layer).

    Note that all this split for :class:`ClientLocal` and :class:`ClientRemote` (and related other classes) is in
    anticipation that there is a way to keep data persisted in offline/local index which allows
    for the server to be embedded in the client (which can only be practical if server start time is negligible).
    Currently, :class:`ClientLocal` is mostly used for simpler testing (by embedding the `LocalServer`):
    *   Run client and server embedded in the same process.
    *   Avoid communication over the network.
    Access to `LocalServer` in tests:
    *   Func `main()` on client side returns `ClientCommandLocal` in case of `ClientLocal`.
    *   `ClientCommandLocal.local_server` is the `LocalServer` instance to inspection in tests.
    """

    def __init__(
        self,
        client_config: ClientConfig,
    ):
        super().__init__(
            client_config,
            ClientCommandFactoryLocal(),
        )
        assert client_config.use_local_requests



from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc

func_id_ = "func_id"
"""
An id (unique string) used by `AbstractDelegator` to distinguish one of many functions it can potentially support.
See also `ReservedPropName.func_id`.
"""

delegator_plugin_instance_id_ = "delegator_plugin_instance_id"
search_control_list_ = "search_control_list"


# TODO_45_75_75_65: Merge `instance_data` into `envelop_payload`:
class FunctionEnvelopeInstanceDataSchema(Schema):
    """
    Schema for :class:`DataEnvelopeSchema.instance_data` of :class:`ReservedEnvelopeClass.ClassFunction`
    """

    class Meta:
        unknown = RAISE
        strict = True

    func_id = fields.String(
        required = True,
    )

    delegator_plugin_instance_id = fields.String(
        required = True,
    )

    search_control_list = fields.List(
        fields.Nested(search_control_desc.dict_schema),
        required = True,
    )


function_envelope_instance_data_desc = TypeDesc(
    dict_schema = FunctionEnvelopeInstanceDataSchema(),
    ref_name = FunctionEnvelopeInstanceDataSchema.__name__,
    dict_example = {
        func_id_: "some_function",
        # TODO: Be able to import `NoopDelegator` and reference it via `__name__`:
        delegator_plugin_instance_id_: "NoopDelegator.default",
        search_control_list_: [
            search_control_desc.dict_example,
        ],
    },
    default_file_path = "",
)
